from docutils.nodes import Element
from sphinx.addnodes import desc_addname
from sphinx.application import Sphinx
from sphinx.domains.std import GenericObject
from sphinx.environment import BuildEnvironment
from sphinx.roles import EmphasizedLiteral


class PathRole(GenericObject):
    def parse_node(self, env: BuildEnvironment, sig: str, signode: Element) -> str:  # type: ignore
        sig = sig.strip()
        if not (sig.startswith("`") and sig.endswith("`")):
            signode += desc_addname(sig, sig)
            return sig
        sig = sig[1:-1]
        nodes, msgs = EmphasizedLiteral()(
            "samp",
            sig,
            sig,
            self.lineno,
            self.state.inliner,
        )
        signode += desc_addname(
            "",
            "",
            *nodes[0].children,
            role=self.name.lower(),
            classes=["color-path-hl"],
        )
        signode += msgs
        return sig


def setup(app: Sphinx):
    app.add_object_type("color-path", "color-path", "%s")
    app.add_directive_to_domain("std", "color-path", PathRole, override=True)

    app.add_object_type("decoration-path", "decoration-path", "%s")
    app.add_directive_to_domain("std", "decoration-path", PathRole, override=True)

    return {
        "version": "1.0.0",
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



from __future__ import annotations

import io
import platform
import sys

import yuio.io
import yuio.term

import pytest
from sybil import Sybil
from sybil.parsers.codeblock import PythonCodeBlockParser
from sybil.parsers.doctest import DocTestParser
from sybil.parsers.rest import SkipParser

if sys.version_info >= (3, 14):

    def _setup(*_args, **_kwargs):
        yuio.term._TTY_SETUP_PERFORMED = True
        yuio.term._TTY_OUTPUT = None
        yuio.term._TTY_INPUT = None
        yuio.term._TERMINAL_THEME = None
        yuio.term._EXPLICIT_COLOR_SUPPORT = yuio.term.ColorSupport.NONE
        yuio.term._COLOR_SUPPORT = yuio.term.ColorSupport.NONE

        yuio.io.setup(term=yuio.term.Term(io.StringIO(), io.StringIO()))

    def _teardown(*_args, **_kwargs):
        yuio.io.restore_streams()
        yuio.io._manager().stop()
        yuio.io._IO_MANAGER = None
        yuio.term._TTY_SETUP_PERFORMED = False
        del yuio.term._TTY_OUTPUT
        del yuio.term._TTY_INPUT
        del yuio.term._TERMINAL_THEME
        del yuio.term._EXPLICIT_COLOR_SUPPORT
        del yuio.term._COLOR_SUPPORT

    pytest_collect_file = Sybil(
        parsers=[
            DocTestParser(),
            PythonCodeBlockParser(),
            SkipParser(),
        ],
        patterns=["*.rst", "*.py"],
        excludes=["yuio/_vendor/*", "yuio/ext/sphinx.py"],
        setup=_setup,
        teardown=_teardown,
    ).pytest()


_PLATFORMS = {"windows", "linux", "darwin"}


def pytest_runtest_setup(item):
    supported_platforms = _PLATFORMS.intersection(
        mark.name for mark in item.iter_markers()
    )
    plat = platform.system().lower()
    if supported_platforms and plat not in supported_platforms:
        pytest.skip(f"cannot run on platform {plat}")


def pytest_ignore_collect(collection_path, config):
    if collection_path.name.endswith("_314.py"):
        return sys.version_info < (3, 14)

    return False



import importlib.util
import io

import yuio.dbg

import pytest

__version__ = "1.1.2.2"


def test_report():
    sio = io.StringIO()
    yuio.dbg.print_report(dest=sio)
    res = sio.getvalue()
    print(res)

    assert "System" in res
    assert "Versions" in res
    assert "Terminal and CLI" in res


def test_custom_collector():
    sio = io.StringIO()
    yuio.dbg.print_report(
        dest=sio,
        settings=yuio.dbg.ReportSettings(
            collectors=[
                yuio.dbg.Report("x Custom Report x", ["x Custom Report Content x"]),
                lambda: yuio.dbg.Report(
                    "x Custom Report 2 x", ["x Custom Report Content 2 x"]
                ),
                lambda: 1 // 0,  # type: ignore
            ]
        ),
    )
    res = sio.getvalue()
    print(res)

    assert "x Custom Report x" in res
    assert "x Custom Report Content x" in res
    assert "x Custom Report 2 x" in res
    assert "x Custom Report Content 2 x" in res
    assert "test.test_dbg.test_custom_collector.<locals>.<lambda>" in res
    assert "ZeroDivisionError:" in res


def test_app_version():
    try:
        from dbg_hooks import main  # type: ignore
    except ImportError:
        pytest.skip("test/test_deps/dbg_hooks is not installed")

    sio = io.StringIO()
    yuio.dbg.print_report(dest=sio, app=main)
    res = sio.getvalue()
    print(res)

    assert "__app__: 1.0.0.1" in res
    assert "dbg_hooks: 1.2.3" in res


def test_packages():
    try:
        import dbg_hooks  # type: ignore
    except ImportError:
        pytest.skip("test/test_deps/dbg_hooks is not installed")

    sio = io.StringIO()
    yuio.dbg.print_report(
        dest=sio,
        settings=yuio.dbg.ReportSettings(
            package=dbg_hooks,
            dependencies=[
                __name__,
                "__main__",
                "foo.bar",
                dbg_hooks,
                "dbg_dep.mod",
                None,  # type: ignore
            ],
        ),
    )
    res = sio.getvalue()
    print(res)

    assert "__main__: unknown" in res
    assert "dbg-dep[opt]: 3.2.1" in res
    assert "dbg_dep.mod: 100.0.0" in res
    assert "dbg_hooks: 1.2.3" in res
    assert "foo.bar: can't collect this item: No module named 'foo'" in res
    assert "test.test_dbg: 1.1.2.2" in res
    assert "yuio: " + yuio.__version__ in res
    assert "TypeError: expected str or ModuleType, got NoneType: None" in res


def test_hooks():
    if not importlib.util.find_spec("dbg_hooks"):
        pytest.skip("test/test_deps/dbg_hooks is not installed")

    sio = io.StringIO()
    yuio.dbg.print_report(dest=sio)
    res = sio.getvalue()
    print(res)

    assert "Collect Env Ok" in res
    assert "Something something" in res
    assert "foo: bar" in res
    assert "collect_env_err" in res
    assert "RuntimeError: something went wrong" in res
    assert "collect_env_missing" in res
    assert "module 'dbg_hooks' has no attribute 'collect_env_missing'" in res



import json

import yuio.color
import yuio.io

if __name__ == "__main__":
    yuio.io.hr(weight=2)
    yuio.io.hr("<c bold>✨ COLOR PRINTING DEMO ✨</c>", weight=0)
    yuio.io.hr(weight=2)

    yuio.io.heading("Message colors")
    yuio.io.success("Success message is bold green.")
    yuio.io.failure("Failure message is bold red.")
    yuio.io.info("Info message is default color.")
    yuio.io.warning("Warning message is yellow.")
    yuio.io.error("Error message is red.")

    yuio.io.heading("Color tags")
    yuio.io.info("This text is <c red bold>rad!</c>")
    yuio.io.warning("File <c code>example.txt</c> does not exist.")

    yuio.io.heading("Links")
    yuio.io.info(
        "Visit %s (if your terminal supports hyperlinks, that is.)",
        yuio.io.Link("example.com", url="https://example.com"),
    )
    yuio.io.info(
        "Or open %s in a text editor.",
        yuio.io.Link.from_path("this example", path=__file__),
    )

    yuio.io.heading("Exceptions")
    try:
        json.loads("{ nah, this is not a valid JSON 😕 }")
    except:
        yuio.io.error_with_tb("Something went horribly wrong!")

    yuio.io.heading("Markdown")
    yuio.io.md(
        """
        - You can use `` `backticks` `` with all functions from `yuio.io`.
        - You can also use CommonMark block markup with `yuio.io.md`.
        - For example, check out this fork bomb:

          ```sh
          :(){ :|:& };:  # <- don't paste this in bash!
          ```
        """
    )

    yuio.io.heading("Code highlighting")
    yuio.io.hl(
        """
        {
            "config": "~/.config.json",
            "syntax": "json"
        }
        """,
        syntax="json",
    )

    yuio.io.heading("Low level API")
    yuio.io.raw(
        yuio.io.Stack(
            yuio.io.Hr(
                "low level API start",
                weight=0,
                left_middle=">",
                left_end=" ",
                right_start=" ",
                right_middle="<",
            ),
            yuio.io.Wrap(
                yuio.io.Format(
                    "Some paragraph that will be wrapped to 40 symbols.\n\n"
                    "This part uses `yuio.io`, `yuio.string`, `yuio.theme` and other "
                    "low-level modules. You can do some %s things with them!",
                    yuio.io.ColorizedString(
                        [
                            yuio.color.Color.STYLE_INVERSE
                            | yuio.color.Color.STYLE_BLINK
                            | yuio.color.Color.FORE_MAGENTA,
                            "advanced",
                        ]
                    ),
                ),
                width=40,
                indent=yuio.io.ColorizedString(
                    [yuio.io.get_theme().get_color("msg/decoration"), "> "]
                ),
            ),
            yuio.io.Hr(
                "low level API end",
                weight=0,
                left_middle=">",
                left_end=" ",
                right_start=" ",
                right_middle="<",
            ),
        ),
        add_newline=True,
    )



import contextlib
import io
import os
import sys

import yuio.color
import yuio.term

import pytest


class MockOStream(io.StringIO):
    def __init__(
        self,
        out: "MockIStream",
        tty: bool,
        should_query_osc: bool,
        osc_response: str | None,
    ):
        super().__init__()
        self.__out = out
        self.__tty = tty
        self.__should_query_osc = should_query_osc
        self.__osc_response = osc_response

    def isatty(self) -> bool:
        return self.__tty

    _OSC_Q = (
        "\x1b]10;?\x1b\\"
        "\x1b]11;?"
        "\x1b\\"
        "\x1b]4;0;?\x1b\\"
        "\x1b]4;1;?\x1b\\"
        "\x1b]4;2;?\x1b\\"
        "\x1b]4;3;?\x1b\\"
        "\x1b]4;4;?\x1b\\"
        "\x1b]4;5;?\x1b\\"
        "\x1b]4;6;?\x1b\\"
        "\x1b]4;7;?\x1b\\"
        "\x1b[c"
    )

    _OSC_R = (
        "\x1b]10;rgb:bfbf/bfbf/bfbf\x1b\\"
        "\x1b]11;rgb:0000/0000/0000\x1b\\"
        "\x1b]4;0;rgb:0000/0000/0000\x1b\\"
        "\x1b]4;1;rgb:d4d4/2c2c/3a3a\x1b\\"
        "\x1b]4;2;rgb:1c1c/a8a8/0000\x1b\\"
        "\x1b]4;3;rgb:c0c0/a0a0/0000\x1b\\"
        "\x1b]4;4;rgb:0000/5d5d/ffff\x1b\\"
        "\x1b]4;5;rgb:b1b1/4848/c6c6\x1b\\"
        "\x1b]4;6;rgb:0000/a8a8/9a9a\x1b\\"
        "\x1b]4;7;rgb:bfbf/bfbf/bfbf\x1b\\"
        "\x1b[?c"
    )

    def write(self, s) -> int:
        if s == self._OSC_Q and self.__osc_response != "":
            if not self.__should_query_osc:
                raise RuntimeError("terminal is not supposed to query OSC")
            self._add_to_out(self.__osc_response or self._OSC_R)
        return super().write(s)

    def writelines(self, lines):
        for line in lines:
            self.write(line)

    def _add_to_out(self, s):
        pos = self.__out.tell()
        self.__out.seek(0, io.SEEK_END)
        self.__out.write(s)
        self.__out.seek(pos)
        self.__out.flush()


class MockIStream(io.StringIO):
    def __init__(
        self,
        tty: bool,
    ):
        super().__init__()
        self.__tty = tty

    def isatty(self) -> bool:
        return self.__tty

    def read(self, size=None):
        raise RuntimeError("term not supposed to use stream.read")

    def readline(self, size=None):
        raise RuntimeError("term not supposed to use stream.readline")

    def readlines(self, hint=-1):
        raise RuntimeError("term not supposed to use stream.readlines")

    def read_keycode(self, *_, **__):
        return super().read()


@contextlib.contextmanager
def mock_term_io(
    i_tty: bool = False,
    o_tty: bool = False,
    should_query_osc: bool = False,
    is_foreground: bool = False,
    env: dict[str, str] = {},
    args: list[str] = [],
    osc_response: str | None = "",
    enable_vt_processing: bool = False,
):
    istream = MockIStream(i_tty)
    ostream = MockOStream(istream, o_tty, should_query_osc, osc_response)

    old_env = os.environ.copy()
    os.environ.clear()
    os.environ.update(env)

    old_args = sys.argv[1:]
    sys.argv[1:] = args

    old_read_keycode, yuio.term._read_keycode = (
        yuio.term._read_keycode,
        istream.read_keycode,
    )
    old_is_foreground, yuio.term._is_foreground = (
        yuio.term._is_foreground,
        lambda *_, **__: is_foreground,
    )
    old_input_is_tty, yuio.term._input_is_tty = (
        yuio.term._input_is_tty,
        lambda *_, **__: i_tty,
    )
    old_output_is_tty, yuio.term._output_is_tty = (
        yuio.term._output_is_tty,
        lambda *_, **__: o_tty,
    )
    old_enter_raw_mode, yuio.term._enter_raw_mode = (
        yuio.term._enter_raw_mode,
        lambda *_, **__: contextlib.nullcontext(),
    )
    old_enable_vt_processing, yuio.term._enable_vt_processing = (
        yuio.term._enable_vt_processing,
        lambda *_, **__: enable_vt_processing,
    )
    old_flush_input_buffer, yuio.term._flush_input_buffer = (
        yuio.term._flush_input_buffer,
        lambda *_, **__: None,
    )

    try:
        yield ostream, istream
    finally:
        os.environ.clear()
        os.environ.update(old_env)

        sys.argv[1:] = old_args

        yuio.term._read_keycode = old_read_keycode
        yuio.term._is_foreground = old_is_foreground
        yuio.term._input_is_tty = old_input_is_tty
        yuio.term._output_is_tty = old_output_is_tty
        yuio.term._enter_raw_mode = old_enter_raw_mode
        yuio.term._enable_vt_processing = old_enable_vt_processing
        yuio.term._flush_input_buffer = old_flush_input_buffer


term_colors = yuio.term.TerminalTheme(
    background=yuio.color.ColorValue.from_hex("#000000"),
    foreground=yuio.color.ColorValue.from_hex("#BFBFBF"),
    black=yuio.color.ColorValue.from_hex("#000000"),
    red=yuio.color.ColorValue.from_hex("#D42C3A"),
    green=yuio.color.ColorValue.from_hex("#1CA800"),
    yellow=yuio.color.ColorValue.from_hex("#C0A000"),
    blue=yuio.color.ColorValue.from_hex("#005DFF"),
    magenta=yuio.color.ColorValue.from_hex("#B148C6"),
    cyan=yuio.color.ColorValue.from_hex("#00A89A"),
    white=yuio.color.ColorValue.from_hex("#BFBFBF"),
    lightness=yuio.term.Lightness.DARK,
)


@pytest.mark.parametrize(
    ("level", "ansi", "ansi_256", "ansi_true"),
    [
        (yuio.color.ColorSupport.ANSI, True, False, False),
        (yuio.color.ColorSupport.ANSI_256, True, True, False),
        (yuio.color.ColorSupport.ANSI_TRUE, True, True, True),
    ],
)
def test_color_support(level, ansi, ansi_256, ansi_true):
    term = yuio.term.Term(None, None, color_support=level)  # type: ignore
    assert term.supports_colors == ansi
    assert term.supports_colors_256 == ansi_256
    assert term.supports_colors_true == ansi_true


# @pytest.mark.linux
# @pytest.mark.darwin
# @pytest.mark.parametrize(
#     ("kwargs", "expected_term"),
#     [
#         (
#             {},
#             {
#                 "color_support": yuio.color.ColorSupport.NONE,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "terminal_theme": None,
#             },
#         ),
#         (
#             {"env": {"TERM": "xterm"}},
#             {
#                 "color_support": yuio.color.ColorSupport.NONE,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "terminal_theme": None,
#             },
#         ),
#         (
#             {
#                 "env": {"TERM": "xterm"},
#                 "i_tty": True,
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.NONE,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.BACKGROUND,
#                 "terminal_theme": None,
#             },
#         ),
#         (
#             {
#                 "env": {"TERM": "xterm"},
#                 "o_tty": True,
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.ANSI,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.BACKGROUND,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "terminal_theme": None,
#             },
#         ),
#         (
#             {
#                 "env": {"TERM": "xterm"},
#                 "i_tty": True,
#                 "o_tty": True,
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.ANSI,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.BACKGROUND,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.BACKGROUND,
#                 "terminal_theme": None,
#             },
#         ),
#         (
#             {
#                 "env": {"TERM": "xterm"},
#                 "is_foreground": True,
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.NONE,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "terminal_theme": None,
#             },
#         ),
#         (
#             {
#                 "env": {"TERM": "xterm"},
#                 "i_tty": True,
#                 "is_foreground": True,
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.NONE,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.INTERACTIVE,
#                 "terminal_theme": None,
#             },
#         ),
#         (
#             {
#                 "env": {"TERM": "xterm"},
#                 "o_tty": True,
#                 "is_foreground": True,
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.ANSI,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.INTERACTIVE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "terminal_theme": None,
#             },
#         ),
#         (
#             {
#                 "env": {"TERM": "xterm"},
#                 "i_tty": True,
#                 "o_tty": True,
#                 "is_foreground": True,
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.ANSI,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.INTERACTIVE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.INTERACTIVE,
#                 "terminal_theme": None,
#             },
#         ),
#         (
#             {
#                 "env": {"TERM": "xterm", "COLORTERM": "yes"},
#                 "i_tty": True,
#                 "o_tty": True,
#                 "is_foreground": True,
#                 "should_query_osc": True,
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.ANSI_256,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.INTERACTIVE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.INTERACTIVE,
#                 "terminal_theme": None,  # OSC query got no response
#             },
#         ),
#         (
#             {
#                 "env": {"TERM": "xterm", "COLORTERM": "truecolor"},
#                 "i_tty": True,
#                 "o_tty": True,
#                 "is_foreground": True,
#                 "should_query_osc": True,
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.ANSI_TRUE,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.INTERACTIVE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.INTERACTIVE,
#                 "terminal_theme": None,  # OSC query got no response
#             },
#         ),
#         (
#             {
#                 "env": {"TERM": "xterm", "COLORTERM": "yes"},
#                 "i_tty": True,
#                 "o_tty": True,
#                 "is_foreground": True,
#                 "should_query_osc": True,
#                 "osc_response": "\x1b[?c",
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.ANSI_256,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.INTERACTIVE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.INTERACTIVE,
#                 "terminal_theme": None,  # kbhit responds, but OSC is not properly supported
#             },
#         ),
#         (
#             {
#                 "env": {"TERM": "xterm", "COLORTERM": "yes"},
#                 "i_tty": True,
#                 "o_tty": True,
#                 "is_foreground": True,
#                 "should_query_osc": True,
#                 "osc_response": None,  # default response
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.ANSI_256,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.INTERACTIVE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.INTERACTIVE,
#                 "terminal_theme": term_colors,  # Got the response!
#             },
#         ),
#         (
#             {
#                 "env": {"TERM": "xterm", "COLORTERM": "yes"},
#                 "i_tty": True,
#                 "o_tty": True,
#                 "is_foreground": True,
#                 "args": ["--no-color"],
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.NONE,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.INTERACTIVE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.INTERACTIVE,
#                 "terminal_theme": None,
#             },
#         ),
#         (
#             {
#                 "env": {"TERM": "xterm", "COLORTERM": "yes", "NO_COLOR": "1"},
#                 "i_tty": True,
#                 "o_tty": True,
#                 "is_foreground": True,
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.NONE,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.INTERACTIVE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.INTERACTIVE,
#                 "terminal_theme": None,
#             },
#         ),
#         (
#             {
#                 "env": {"TERM": "xterm", "COLORTERM": "truecolor"},
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.NONE,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "terminal_theme": None,
#             },
#         ),
#         (
#             {
#                 "args": ["--color"],
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.ANSI,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "terminal_theme": None,
#             },
#         ),
#         (
#             {
#                 "args": ["--color"],
#                 "env": {"TERM": "xterm", "COLORTERM": "truecolor"},
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.ANSI_TRUE,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "terminal_theme": None,
#             },
#         ),
#         (
#             {"env": {"FORCE_COLOR": "1"}},
#             {
#                 "color_support": yuio.color.ColorSupport.ANSI,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "terminal_theme": None,
#             },
#         ),
#         (
#             {
#                 "env": {"TERM": "xterm", "COLORTERM": "yes"},
#                 "i_tty": True,
#                 "o_tty": True,
#                 "is_foreground": True,
#                 "args": ["--color=0"],
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.NONE,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.INTERACTIVE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.INTERACTIVE,
#                 "terminal_theme": None,
#             },
#         ),
#         (
#             {
#                 "args": ["--color=1"],
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.ANSI,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "terminal_theme": None,
#             },
#         ),
#         (
#             {
#                 "args": ["--color=ansi"],
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.ANSI,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "terminal_theme": None,
#             },
#         ),
#         (
#             {
#                 "args": ["--color=ansi-256"],
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.ANSI_256,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "terminal_theme": None,
#             },
#         ),
#         (
#             {
#                 "args": ["--color=ansi-true"],
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.ANSI_TRUE,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "terminal_theme": None,
#             },
#         ),
#     ],
# )
# def test_capabilities_estimation(kwargs, expected_term):
#     with mock_term_io(**kwargs) as streams:
#         ostream, istream = streams
#         term = yuio.term.get_term_from_stream(ostream, istream)
#         expected = yuio.term.Term(ostream, istream, **expected_term)
#         assert term == expected


# @pytest.mark.windows
# @pytest.mark.parametrize(
#     ("kwargs", "expected_term"),
#     [
#         (
#             {},
#             {
#                 "color_support": yuio.color.ColorSupport.NONE,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "terminal_theme": None,
#             },
#         ),
#         (
#             {"enable_vt_processing": True},
#             {
#                 "color_support": yuio.color.ColorSupport.NONE,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "terminal_theme": None,
#             },
#         ),
#         (
#             {
#                 "i_tty": True,
#                 "enable_vt_processing": True,
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.NONE,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.BACKGROUND,
#                 "terminal_theme": None,
#             },
#         ),
#         (
#             {
#                 "o_tty": True,
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.NONE,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.BACKGROUND,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "terminal_theme": None,
#             },
#         ),
#         (
#             {
#                 "o_tty": True,
#                 "enable_vt_processing": True,
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.ANSI_TRUE,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.BACKGROUND,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "terminal_theme": None,
#             },
#         ),
#         (
#             {
#                 "i_tty": True,
#                 "o_tty": True,
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.NONE,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.BACKGROUND,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.BACKGROUND,
#                 "terminal_theme": None,
#             },
#         ),
#         (
#             {
#                 "i_tty": True,
#                 "o_tty": True,
#                 "enable_vt_processing": True,
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.ANSI_TRUE,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.BACKGROUND,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.BACKGROUND,
#                 "terminal_theme": None,
#             },
#         ),
#         (
#             {
#                 "is_foreground": True,
#                 "enable_vt_processing": True,
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.NONE,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "terminal_theme": None,
#             },
#         ),
#         (
#             {
#                 "i_tty": True,
#                 "is_foreground": True,
#                 "enable_vt_processing": True,
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.NONE,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.INTERACTIVE,
#                 "terminal_theme": None,
#             },
#         ),
#         (
#             {
#                 "o_tty": True,
#                 "is_foreground": True,
#                 "enable_vt_processing": True,
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.ANSI_TRUE,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.INTERACTIVE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "terminal_theme": None,
#             },
#         ),
#         (
#             {
#                 "i_tty": True,
#                 "o_tty": True,
#                 "is_foreground": True,
#                 "enable_vt_processing": True,
#                 "should_query_osc": True,
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.ANSI_TRUE,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.INTERACTIVE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.INTERACTIVE,
#                 "terminal_theme": None,  # OSC query got no response
#             },
#         ),
#         (
#             {
#                 "i_tty": True,
#                 "o_tty": True,
#                 "is_foreground": True,
#                 "enable_vt_processing": True,
#                 "should_query_osc": True,
#                 "osc_response": "\x1b[?c",
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.ANSI_TRUE,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.INTERACTIVE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.INTERACTIVE,
#                 "terminal_theme": None,  # kbhit responds, but OSC is not properly supported
#             },
#         ),
#         (
#             {
#                 "i_tty": True,
#                 "o_tty": True,
#                 "is_foreground": True,
#                 "enable_vt_processing": True,
#                 "should_query_osc": True,
#                 "osc_response": None,  # default response
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.ANSI_TRUE,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.INTERACTIVE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.INTERACTIVE,
#                 "terminal_theme": term_colors,  # Got the response!
#             },
#         ),
#         (
#             {
#                 "i_tty": True,
#                 "o_tty": True,
#                 "is_foreground": True,
#                 "enable_vt_processing": True,
#                 "args": ["--no-color"],
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.NONE,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.INTERACTIVE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.INTERACTIVE,
#                 "terminal_theme": None,
#             },
#         ),
#         (
#             {
#                 "env": {"NO_COLOR": "1"},
#                 "i_tty": True,
#                 "o_tty": True,
#                 "is_foreground": True,
#                 "enable_vt_processing": True,
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.NONE,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.INTERACTIVE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.INTERACTIVE,
#                 "terminal_theme": None,
#             },
#         ),
#         (
#             {
#                 "args": ["--color"],
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.ANSI,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "terminal_theme": None,
#             },
#         ),
#         (
#             {"env": {"FORCE_COLOR": "1"}},
#             {
#                 "color_support": yuio.color.ColorSupport.ANSI,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "terminal_theme": None,
#             },
#         ),
#         (
#             {
#                 "i_tty": True,
#                 "o_tty": True,
#                 "is_foreground": True,
#                 "enable_vt_processing": True,
#                 "args": ["--color=0"],
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.NONE,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.INTERACTIVE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.INTERACTIVE,
#                 "terminal_theme": None,
#             },
#         ),
#         (
#             {
#                 "args": ["--color=1"],
#             },
#             {
#                 "color_support": yuio.color.ColorSupport.ANSI,
#                 "ostream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "istream_interactive_support": yuio.term.InteractiveSupport.NONE,
#                 "terminal_theme": None,
#             },
#         ),
#     ],
# )
# def test_capabilities_estimation_windows(kwargs, expected_term):
#     with mock_term_io(**kwargs) as streams:
#         ostream, istream = streams
#         term = yuio.term.get_term_from_stream(ostream, istream)
#         expected = yuio.term.Term(ostream, istream, **expected_term)
#         assert term == expected


@pytest.mark.parametrize(
    ("env_vars", "expected"),
    [
        ({"GITHUB_ACTIONS": ""}, yuio.color.ColorSupport.ANSI_TRUE),
        ({"TRAVIS": ""}, yuio.color.ColorSupport.ANSI),
        ({"CIRCLECI": ""}, yuio.color.ColorSupport.ANSI),
        ({"APPVEYOR": ""}, yuio.color.ColorSupport.ANSI),
        ({"GITLAB_CI": ""}, yuio.color.ColorSupport.ANSI),
        ({"BUILDKITE": ""}, yuio.color.ColorSupport.ANSI),
        ({"DRONE": ""}, yuio.color.ColorSupport.ANSI),
        ({"TEAMCITY_VERSION": "2023.1"}, yuio.color.ColorSupport.ANSI),
        ({}, yuio.color.ColorSupport.NONE),
        ({"PATH": "/usr/bin", "HOME": "/home/user"}, yuio.color.ColorSupport.NONE),
        (
            {"GITHUB_ACTIONS": "", "TRAVIS": ""},
            yuio.color.ColorSupport.ANSI_TRUE,
        ),
        ({"TRAVIS": "", "CIRCLECI": ""}, yuio.color.ColorSupport.ANSI),
        ({"CI": ""}, yuio.color.ColorSupport.NONE),
    ],
)
def test_detect_ci_color_support(env_vars, expected, monkeypatch: pytest.MonkeyPatch):
    for k, v in env_vars.items():
        monkeypatch.setenv(k, v)
    assert yuio.term.detect_ci_color_support() == expected



import io
import sys
import textwrap

import yuio.io
import yuio.parse
import yuio.string
import yuio.term
import yuio.theme
import yuio.widget
from yuio import _t

from .conftest import IOMocker, RcCompare

import pytest


class TestSetup:
    @pytest.fixture(autouse=True)
    def setup_io_fresh(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr("yuio.io._IO_MANAGER", None)

        yield

        yuio.io.restore_streams()
        if yuio.io._IO_MANAGER is not None:
            yuio.io._IO_MANAGER.stop()
            yuio.io._IO_MANAGER = None

    def test_implicit(self, monkeypatch: pytest.MonkeyPatch):
        stdout = io.StringIO()
        monkeypatch.setattr("sys.stdout", stdout)

        stderr = io.StringIO()
        monkeypatch.setattr("sys.stderr", stderr)

        stdin = io.StringIO()
        monkeypatch.setattr("sys.stdin", stdin)

        assert yuio.io.get_term().ostream is stderr
        assert yuio.io.get_term().istream is stdin

        assert yuio.io.orig_stderr() is stderr
        assert sys.stderr is stderr
        assert sys.stdin is stdin
        assert not yuio.io.streams_wrapped()

        yuio.io.info("Foo bar!")

        assert stderr.getvalue() == "Foo bar!\n"

    def test_term(self, term):
        yuio.io.setup(term=term)
        assert yuio.io.get_term() is term
        assert isinstance(yuio.io.get_theme(), yuio.theme.DefaultTheme)

    def test_theme(self, term):
        theme = yuio.theme.DefaultTheme(term)
        yuio.io.setup(theme=theme)
        assert yuio.io.get_theme() is theme

    def test_theme_callable(self, term):
        n_called = 0
        current_theme = None
        current_term = None

        def theme(term):
            nonlocal n_called, current_theme, current_term
            n_called += 1
            current_theme = yuio.theme.Theme()
            current_term = term
            return current_theme

        yuio.io.setup(theme=theme)
        assert n_called == 1
        assert yuio.io.get_theme() is current_theme
        assert yuio.io.get_term() is current_term

        yuio.io.setup(theme=theme)
        assert n_called == 2
        assert yuio.io.get_theme() is current_theme
        assert yuio.io.get_term() is current_term

        yuio.io.setup(term=term)
        assert n_called == 3
        assert current_term is term
        assert yuio.io.get_theme() is current_theme
        assert yuio.io.get_term() is current_term

    def test_wrap_streams(self, monkeypatch: pytest.MonkeyPatch, ostream):
        monkeypatch.setattr("sys.stdout", ostream)
        monkeypatch.setattr("sys.stderr", ostream)

        yuio.io.setup(wrap_stdio=True)

        assert yuio.io.streams_wrapped()
        assert sys.stdout is not ostream
        assert getattr(sys.stdout, "_WrappedOutput__wrapped") is ostream
        assert yuio.io.orig_stderr() is ostream
        assert sys.stderr is not ostream
        assert getattr(sys.stderr, "_WrappedOutput__wrapped") is ostream
        assert yuio.io.orig_stdout() is ostream

        yuio.io.restore_streams()
        assert not yuio.io.streams_wrapped()
        assert sys.stdout is ostream
        assert sys.stderr is ostream
        assert yuio.io.orig_stdout() is ostream
        assert yuio.io.orig_stderr() is ostream

        yuio.io.restore_streams()
        assert not yuio.io.streams_wrapped()
        assert sys.stdout is ostream
        assert sys.stderr is ostream
        assert yuio.io.orig_stdout() is ostream
        assert yuio.io.orig_stderr() is ostream

    def test_wrap_streams_not_tty(self, monkeypatch: pytest.MonkeyPatch, ostream):
        monkeypatch.setattr(ostream, "isatty", lambda: False)
        monkeypatch.setattr("sys.stdout", ostream)
        monkeypatch.setattr("sys.stderr", ostream)

        yuio.io.setup(wrap_stdio=True)

        assert yuio.io.streams_wrapped()
        assert sys.stdout is ostream
        assert yuio.io.orig_stderr() is ostream
        assert sys.stderr is ostream
        assert yuio.io.orig_stdout() is ostream

        yuio.io.restore_streams()
        assert not yuio.io.streams_wrapped()
        assert sys.stdout is ostream
        assert sys.stderr is ostream
        assert yuio.io.orig_stdout() is ostream
        assert yuio.io.orig_stderr() is ostream

        yuio.io.restore_streams()
        assert not yuio.io.streams_wrapped()
        assert sys.stdout is ostream
        assert sys.stderr is ostream
        assert yuio.io.orig_stdout() is ostream
        assert yuio.io.orig_stderr() is ostream


class TestMessage:
    def test_info(self, ostream: io.StringIO):
        yuio.io.info("foo bar!")
        assert RcCompare.from_commands(ostream.getvalue()) == RcCompare(
            [
                "foo bar!            ",
            ],
            [
                "cccccccc            ",
            ],
        )

    def test_info_wrap(self, ostream: io.StringIO):
        yuio.io.info("foo bar baz foo bar baz foo bar baz!")
        assert RcCompare.from_commands(ostream.getvalue()) == RcCompare(
            [
                "foo bar baz foo bar ",
                "baz foo bar baz!    ",
            ],
            [
                "cccccccccccccccccccc",
                "cccccccccccccccc    ",
            ],
        )

    def test_info_args(self, ostream: io.StringIO):
        yuio.io.info("%s %s!", "foo", "bar")
        assert RcCompare.from_commands(ostream.getvalue()) == RcCompare(
            [
                "foo bar!            ",
            ],
            [
                "cccccccc            ",
            ],
        )

    def test_info_color_tags(self, ostream: io.StringIO):
        yuio.io.info("`%s` <c bold>%s</c>!", "<c green>foo</c>", "bar")
        assert RcCompare.from_commands(ostream.getvalue()) == RcCompare(
            [
                "<c green>foo</c> bar",
                "!                   ",
            ],
            [
                "mmmmmmmmmmmmmmmmcCCC",
                "c                   ",
            ],
        )

    def test_warning(self, ostream: io.StringIO):
        yuio.io.warning("foo bar!")
        assert RcCompare.from_commands(ostream.getvalue()) == RcCompare(
            [
                "foo bar!            ",
            ],
            [
                "yyyyyyyy            ",
            ],
        )

    def test_error(self, ostream: io.StringIO):
        yuio.io.error("foo bar!")
        assert RcCompare.from_commands(ostream.getvalue()) == RcCompare(
            [
                "foo bar!            ",
            ],
            [
                "rrrrrrrr            ",
            ],
        )

    def test_error_with_tb(self, ostream: io.StringIO, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr(
            "traceback.format_exception", lambda ty, val, tb: [f"{ty.__name__}: {val}"]
        )

        try:
            raise RuntimeError("something happened")
        except RuntimeError:
            yuio.io.error_with_tb("foo bar!")
        assert RcCompare.from_commands(ostream.getvalue()) == RcCompare(
            [
                "foo bar!            ",
                "  RuntimeError: some",
                "thing happened      ",
            ],
            [
                "rrrrrrrr            ",
                "                    ",
                "                    ",
            ],
        )

    def test_error_with_tb_manual(
        self, ostream: io.StringIO, monkeypatch: pytest.MonkeyPatch
    ):
        monkeypatch.setattr(
            "traceback.format_exception", lambda ty, val, tb: [f"{ty.__name__}: {val}"]
        )

        try:
            raise RuntimeError("something happened")
        except RuntimeError as e:
            yuio.io.error_with_tb("foo bar!", exc_info=(TypeError, e, e.__traceback__))
        assert RcCompare.from_commands(ostream.getvalue()) == RcCompare(
            [
                "foo bar!            ",
                "  TypeError: somethi",
                "ng happened         ",
            ],
            [
                "rrrrrrrr            ",
                "                    ",
                "                    ",
            ],
        )

    def test_failure(self, ostream: io.StringIO):
        yuio.io.failure("foo bar!")
        assert RcCompare.from_commands(ostream.getvalue()) == RcCompare(
            [
                "foo bar!            ",
            ],
            [
                "RRRRRRRR            ",
            ],
        )

    def test_failure_with_tb(
        self, ostream: io.StringIO, monkeypatch: pytest.MonkeyPatch
    ):
        monkeypatch.setattr(
            "traceback.format_exception", lambda ty, val, tb: [f"{ty.__name__}: {val}"]
        )

        try:
            raise RuntimeError("something happened")
        except RuntimeError:
            yuio.io.failure_with_tb("foo bar!")
        assert RcCompare.from_commands(ostream.getvalue()) == RcCompare(
            [
                "foo bar!            ",
                "  RuntimeError: some",
                "thing happened      ",
            ],
            [
                "RRRRRRRR            ",
                "                    ",
                "                    ",
            ],
        )

    def test_failure_with_tb_manual(
        self, ostream: io.StringIO, monkeypatch: pytest.MonkeyPatch
    ):
        monkeypatch.setattr(
            "traceback.format_exception", lambda ty, val, tb: [f"{ty.__name__}: {val}"]
        )

        try:
            raise RuntimeError("something happened")
        except RuntimeError as e:
            yuio.io.failure_with_tb(
                "foo bar!", exc_info=(TypeError, e, e.__traceback__)
            )
        assert RcCompare.from_commands(ostream.getvalue()) == RcCompare(
            [
                "foo bar!            ",
                "  TypeError: somethi",
                "ng happened         ",
            ],
            [
                "RRRRRRRR            ",
                "                    ",
                "                    ",
            ],
        )

    def test_failure_with_tb_manual_2(
        self, ostream: io.StringIO, monkeypatch: pytest.MonkeyPatch
    ):
        monkeypatch.setattr(
            "traceback.format_exception", lambda ty, val, tb: [f"{ty.__name__}: {val}"]
        )

        try:
            raise RuntimeError("something happened")
        except RuntimeError as e:
            yuio.io.failure_with_tb("foo bar!", exc_info=e)
        assert RcCompare.from_commands(ostream.getvalue()) == RcCompare(
            [
                "foo bar!            ",
                "  RuntimeError: some",
                "thing happened      ",
            ],
            [
                "RRRRRRRR            ",
                "                    ",
                "                    ",
            ],
        )

    def test_success(self, ostream: io.StringIO):
        yuio.io.success("foo bar!")
        assert RcCompare.from_commands(ostream.getvalue()) == RcCompare(
            [
                "foo bar!            ",
            ],
            [
                "gggggggg            ",
            ],
        )

    def test_heading(self, ostream: io.StringIO):
        yuio.io.heading("foo bar!")
        yuio.io.heading("foo bar 2!")
        yuio.io.heading("foo bar 3!", wrap=False)
        assert RcCompare.from_commands(ostream.getvalue()) == RcCompare(
            [
                "⣿ foo bar!          ",
                "                    ",
                "⣿ foo bar 2!        ",
                "                    ",
                "⣿ foo bar 3!        ",
                "                    ",
            ],
            [
                "mm########          ",
                "                    ",
                "mm##########        ",
                "                    ",
                "mm##########        ",
                "                    ",
            ],
        )

    def test_md(self, ostream: io.StringIO):
        yuio.io.md("# Foo!\n\n- bar\n- baz")
        assert RcCompare.from_commands(ostream.getvalue()) == RcCompare(
            [
                "⣿ Foo!              ",
                "                    ",
                "•   bar             ",
                "                    ",
                "•   baz             ",
            ],
            [
                "mmCCCC              ",
                "                    ",
                "mmmmccc             ",
                "                    ",
                "mmmmccc             ",
            ],
        )

    def test_br(self, ostream: io.StringIO):
        yuio.io.info("foo")
        yuio.io.br()
        yuio.io.info("bar")
        assert RcCompare.from_commands(ostream.getvalue()) == RcCompare(
            [
                "foo                 ",
                "                    ",
                "bar                 ",
            ],
            [
                "ccc                 ",
                "                    ",
                "ccc                 ",
            ],
        )

    def test_hr(self, ostream: io.StringIO):
        yuio.io.hr()
        yuio.io.hr(weight=2)
        yuio.io.hr("foo")
        assert RcCompare.from_commands(ostream.getvalue()) == RcCompare(
            [
                "────────────────────",
                "━━━━━━━━━━━━━━━━━━━━",
                "───────╴foo╶────────",
            ],
            [
                "mmmmmmmmmmmmmmmmmmmm",
                "mmmmmmmmmmmmmmmmmmmm",
                "mmmmmmmmcccmmmmmmmmm",
            ],
        )

    def test_hl(self, ostream):
        yuio.io.hl('{"foo": true}', syntax="json")
        assert RcCompare.from_commands(ostream.getvalue()) == RcCompare(
            [
                '{"foo": true}       ',
            ],
            [
                "ccccccccccccc       ",
            ],
        )

    def test_raw_err(self):
        with pytest.raises(
            TypeError, match=r"term, to_stdout, to_stderr can't be given together"
        ):
            yuio.io.raw("asd", to_stdout=True, to_stderr=True)

        with pytest.raises(ValueError, match=r"invalid exc_info"):
            yuio.io.raw("asd", exc_info=(1, 2, 3, 4))  # type: ignore


class TestAsk:
    def test_simple(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Hello?              ",
                ">                   ",
                "f1 help             ",
                "                    ",
                "                    ",
            ],
        )
        io_mocker.text("Hii~")
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            assert yuio.io.ask("Hello?") == "Hii~"

    def test_empty(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Hello?              ",
                ">                   ",
                "f1 help             ",
                "                    ",
                "                    ",
            ],
        )
        io_mocker.key(yuio.widget.Key.ENTER)
        io_mocker.expect_screen(
            [
                "Hello?              ",
                ">                   ",
                "▲ Input is required ",
                "f1 help             ",
                "                    ",
            ],
        )
        io_mocker.text("123")
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            assert yuio.io.ask("Hello?") == "123"

    def test_format(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "What's your deal?   ",
                ">                   ",
                "f1 help             ",
                "                    ",
                "                    ",
            ],
        )
        io_mocker.text("meow =^..^=")
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            assert yuio.io.ask("What's your %s?", "deal") == "meow =^..^="

    def test_add_colon(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Enter something:    ",
                ">                   ",
                "f1 help             ",
                "                    ",
                "                    ",
            ],
        )
        io_mocker.text("123")
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            assert yuio.io.ask("Enter something") == "123"

    def test_dont_add_colon(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Enter something:    ",
                ">                   ",
                "f1 help             ",
                "                    ",
                "                    ",
            ],
        )
        io_mocker.text("123")
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            assert yuio.io.ask("Enter something:") == "123"

    def test_default(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Q?                  ",
                "> {default}         ",
                "f1 help             ",
                "                    ",
                "                    ",
            ],
        )
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            assert yuio.io.ask("Q?", default="{default}") == "{default}"

    def test_default_overridden(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Q?                  ",
                "> {default}         ",
                "f1 help             ",
                "                    ",
                "                    ",
            ],
        )
        io_mocker.text("foo!")
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            assert yuio.io.ask("Q?", default="{default}") == "foo!"

    def test_default_optional(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Q?                  ",
                "> <none>            ",
                "f1 help             ",
                "                    ",
                "                    ",
            ],
        )
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            assert yuio.io.ask("Q?", default=None) is None

    def test_default_other_type(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Q?                  ",
                "> 10                ",
                "f1 help             ",
                "                    ",
                "                    ",
            ],
        )
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            assert yuio.io.ask("Q?", default=10) == 10

    def test_default_add_colon(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Q:                  ",
                "> {default}         ",
                "f1 help             ",
                "                    ",
                "                    ",
            ],
        )
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            assert yuio.io.ask("Q", default="{default}") == "{default}"

    def test_default_dont_add_colon(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Q:                  ",
                "> {default}         ",
                "f1 help             ",
                "                    ",
                "                    ",
            ],
        )
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            assert yuio.io.ask("Q:", default="{default}") == "{default}"

    def test_input_description(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Q:                  ",
                ">                   ",
                "<text> • f1 help    ",
                "                    ",
                "                    ",
            ],
        )
        io_mocker.key(yuio.widget.Key.F1)
        io_mocker.expect_screen(
            [
                "Input Format        ",
                "                    ",
                "            <text>  ",
                "                    ",
                ":                   ",
            ],
        )
        io_mocker.key(yuio.widget.Key.ESCAPE)
        io_mocker.text("123")
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            assert yuio.io.ask("Q:", input_description="<text>") == "123"

    def test_default_description(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Q:                  ",
                "> <default>         ",
                "f1 help             ",
                "                    ",
                "                    ",
            ],
        )
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            assert (
                yuio.io.ask("Q:", default="{default}", default_description="<default>")
                == "{default}"
            )

    def test_parser(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Are you there?      ",
                "> no                ",
                "  yes               ",
                "f1 help             ",
                "                    ",
            ],
        )
        io_mocker.key(yuio.widget.Key.TAB)
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            assert yuio.io.ask("Are you there?", parser=yuio.parse.Bool())

    def test_type_hint(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Are you there?      ",
                "> no                ",
                "  yes               ",
                "f1 help             ",
                "                    ",
            ],
        )
        io_mocker.key(yuio.widget.Key.TAB)
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            assert yuio.io.ask[bool]("Are you there?")

    def test_type_hint_and_parser(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Enter some numbers: ",
                ">                   ",
                "f1 help             ",
                "                    ",
                "                    ",
            ],
        )
        io_mocker.text("123 456")
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            assert yuio.io.ask[bool](
                "Enter some numbers",
                parser=yuio.parse.List(yuio.parse.Int()),  # type: ignore
            ) == [123, 456]

    def test_secret(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Enter password:     ",
                ">                   ",
                "f1 help             ",
                "                    ",
                "                    ",
            ],
        )
        io_mocker.text("123 456")
        io_mocker.expect_screen(
            [
                "Enter password:     ",
                "> *******           ",
                "f1 help             ",
                "                    ",
                "                    ",
            ],
        )
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            assert (
                yuio.io.ask[yuio.parse.SecretString]("Enter password").data == "123 456"
            )

    def test_error_highlighting(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Enter some numbers: ",
                ">                   ",
                "f1 help             ",
                "                    ",
                "                    ",
            ],
        )
        io_mocker.text("123 xxx 456")
        io_mocker.key(yuio.widget.Key.ENTER)
        io_mocker.expect_screen(
            [
                "Enter some numbers: ",
                "> 123 xxx 456       ",
                "▲ Can't parse 'xxx' ",
                "as int              ",
                "f1 help             ",
            ],
            [
                "bbbbbbbbbbbbbbbbbbb ",
                "      RRR           ",
                "mrrrrrrrrrrrrrmmmmm ",
                "rrrmmm              ",
                "                    ",
            ],
        )
        io_mocker.key("g", ctrl=True)
        io_mocker.key("w", ctrl=True)
        io_mocker.expect_screen(
            [
                "Enter some numbers: ",
                "> 123  456          ",
                "f1 help             ",
                "                    ",
                "                    ",
            ],
            [
                "bbbbbbbbbbbbbbbbbbb ",
                "                    ",
                "                    ",
                "                    ",
                "                    ",
            ],
        )
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            assert yuio.io.ask[list[int]]("Enter some numbers") == [123, 456]


class TestAskNoColor:
    @pytest.fixture
    def term(self, ostream: io.StringIO, istream: _t.TextIO) -> yuio.term.Term:
        return yuio.term.Term(
            ostream,
            istream,
            color_support=yuio.term.ColorSupport.NONE,
            ostream_is_tty=True,
            istream_is_tty=True,
            is_unicode=True,
        )

    def test_simple(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "> Hello?            ",
            ],
        )
        io_mocker.expect_istream_readline("Hii~")

        with io_mocker.mock():
            assert yuio.io.ask("Hello?") == "Hii~"

    def test_empty(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "> Hello?            ",
            ],
        )
        io_mocker.expect_istream_readline("\n")
        io_mocker.expect_screen(
            [
                "> Hello?            ",
                "Input is required.  ",
                "> Hello?            ",
            ],
        )
        io_mocker.expect_istream_readline("Hii~\n")

        with io_mocker.mock():
            assert yuio.io.ask("Hello?") == "Hii~"

    def test_format(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "> What's your deal? ",
            ],
        )
        io_mocker.expect_istream_readline("meow =^..^=\n")

        with io_mocker.mock():
            assert yuio.io.ask("What's your %s?", "deal") == "meow =^..^="

    def test_add_colon(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "> Enter something:  ",
            ],
        )
        io_mocker.expect_istream_readline("123\n")

        with io_mocker.mock():
            assert yuio.io.ask("Enter something") == "123"

    def test_dont_add_colon(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "> Enter something:  ",
            ],
        )
        io_mocker.expect_istream_readline("123\n")

        with io_mocker.mock():
            assert yuio.io.ask("Enter something:") == "123"

    def test_default(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "> Q? [{default}]    ",
            ],
        )
        io_mocker.expect_istream_readline("\n")

        with io_mocker.mock():
            assert yuio.io.ask("Q?", default="{default}") == "{default}"

    def test_default_overridden(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "> Q? [{default}]    ",
            ],
        )
        io_mocker.expect_istream_readline("foo!\n")

        with io_mocker.mock():
            assert yuio.io.ask("Q?", default="{default}") == "foo!"

    def test_default_optional(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "> Q? [<none>]       ",
            ],
        )
        io_mocker.expect_istream_readline("\n")

        with io_mocker.mock():
            assert yuio.io.ask("Q?", default=None) is None

    def test_default_other_type(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "> Q? [10]           ",
            ],
        )
        io_mocker.expect_istream_readline("\n")

        with io_mocker.mock():
            assert yuio.io.ask("Q?", default=10) == 10

    def test_default_add_colon(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "> Q [{default}]:    ",
            ],
        )
        io_mocker.expect_istream_readline("\n")

        with io_mocker.mock():
            assert yuio.io.ask("Q", default="{default}") == "{default}"

    def test_default_dont_add_colon(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "> Q [{default}]:    ",
            ],
        )
        io_mocker.expect_istream_readline("\n")

        with io_mocker.mock():
            assert yuio.io.ask("Q:", default="{default}") == "{default}"

    def test_input_description(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "> Q (<text>):       ",
            ],
        )
        io_mocker.expect_istream_readline("123\n")

        with io_mocker.mock():
            assert yuio.io.ask("Q:", input_description="<text>") == "123"

    def test_default_description(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "> Q [<default>]:    ",
            ],
        )
        io_mocker.expect_istream_readline("\n")

        with io_mocker.mock():
            assert (
                yuio.io.ask("Q:", default="{default}", default_description="<default>")
                == "{default}"
            )

    def test_parser(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "> Are you there? (  ",
                "  {yes|no})         ",
            ],
        )
        io_mocker.expect_istream_readline("what?\n")
        io_mocker.expect_screen(
            [
                "> Are you there? (  ",
                "  {yes|no}) what?   ",
                "Can't parse 'what?' ",
                "as bool, should be  ",
                "yes, no, true, or   ",
                "false               ",
                "> Are you there? (  ",
                "  {yes|no})         ",
            ],
        )
        io_mocker.expect_istream_readline("y\n")

        with io_mocker.mock():
            assert yuio.io.ask("Are you there?", parser=yuio.parse.Bool())

    def test_parser_hint(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "> Are you there? (  ",
                "  {yes|no})         ",
            ],
        )
        io_mocker.expect_istream_readline("what?\n")
        io_mocker.expect_screen(
            [
                "> Are you there? (  ",
                "  {yes|no}) what?   ",
                "Can't parse 'what?' ",
                "as bool, should be  ",
                "yes, no, true, or   ",
                "false               ",
                "> Are you there? (  ",
                "  {yes|no})         ",
            ],
        )
        io_mocker.expect_istream_readline("y\n")

        with io_mocker.mock():
            assert yuio.io.ask[bool]("Are you there?")

    def test_type_hint_and_parser(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "> Enter some numbers",
                "   (<int>[ <int>[   ",
                "  ...]]):           ",
            ],
        )
        io_mocker.expect_istream_readline("123 456\n")

        with io_mocker.mock():
            assert yuio.io.ask[bool](
                "Enter some numbers",
                parser=yuio.parse.List(yuio.parse.Int()),  # type: ignore
            ) == [123, 456]

    @pytest.mark.linux
    @pytest.mark.darwin
    def test_secret_fallback(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Warning: Password   ",
                "input may be echoed.",
                "> Enter password:   ",
                "                    ",
                "                    ",
            ],
        )
        io_mocker.expect_istream_readline("123 456\n")

        with io_mocker.mock():
            assert (
                yuio.io.ask[yuio.parse.SecretString]("Enter password").data == "123 456"
            )

    @pytest.mark.linux
    @pytest.mark.darwin
    def test_secret_unix(
        self, io_mocker: IOMocker, istream, monkeypatch: pytest.MonkeyPatch
    ):
        import termios

        def fileno():
            return 123

        mode = orig_mode = [0, 0, termios.ECHO, 0, 0, 0, []]

        def tcgetattr(fd):
            assert fd == 123
            return mode[:]

        def tcsetattr(fd, flags, new_mode):
            assert fd == 123
            nonlocal mode
            mode = new_mode[:]

        monkeypatch.setattr(istream, "fileno", fileno)
        monkeypatch.setattr(termios, "tcgetattr", tcgetattr)
        monkeypatch.setattr(termios, "tcsetattr", tcsetattr)

        io_mocker.expect_screen(
            [
                "> Enter password:   ",
                "                    ",
                "                    ",
                "                    ",
                "                    ",
            ],
        )
        io_mocker.expect_eq(lambda: mode[3], 0)
        io_mocker.expect_istream_readline("123 456\n", echo=False)
        io_mocker.expect_mark()
        io_mocker.expect_eq(lambda: mode, orig_mode)
        io_mocker.expect_screen(
            [
                "> Enter password:   ",
                "                    ",
                "                    ",
                "                    ",
                "                    ",
            ],
        )

        with io_mocker.mock():
            assert (
                yuio.io.ask[yuio.parse.SecretString]("Enter password").data == "123 456"
            )
            io_mocker.mark()

    @pytest.mark.linux
    @pytest.mark.darwin
    def test_secret_unix_fallback_if_termios_throws(
        self, io_mocker: IOMocker, istream, monkeypatch: pytest.MonkeyPatch
    ):
        import termios

        def fileno():
            return 123

        def tcgetattr(fd):
            assert fd == 123
            return [0, 0, termios.ECHO, 0, 0, 0, []]

        def tcsetattr(fd, flags, new_mode):
            assert fd == 123
            raise termios.error("error")

        monkeypatch.setattr(istream, "fileno", fileno)
        monkeypatch.setattr(termios, "tcgetattr", tcgetattr)
        monkeypatch.setattr(termios, "tcsetattr", tcsetattr)

        io_mocker.expect_screen(
            [
                "Warning: Password   ",
                "input may be echoed.",
                "> Enter password:   ",
                "                    ",
                "                    ",
            ],
        )
        io_mocker.expect_istream_readline("123 456\n", echo=False)

        with io_mocker.mock():
            assert (
                yuio.io.ask[yuio.parse.SecretString]("Enter password").data == "123 456"
            )

    @pytest.mark.linux
    @pytest.mark.darwin
    def test_secret_unix_raise_if_termios_throws_after_readline(
        self, io_mocker: IOMocker, istream, monkeypatch: pytest.MonkeyPatch
    ):
        import termios

        def fileno():
            return 123

        mode = orig_mode = [0, 0, termios.ECHO, 0, 0, 0, []]
        n_calls = 0

        def tcgetattr(fd):
            assert fd == 123
            return mode[:]

        def tcsetattr(fd, flags, new_mode):
            assert fd == 123
            nonlocal mode, n_calls
            mode = new_mode[:]
            if n_calls == 1:
                raise termios.error("error")
            n_calls += 1

        monkeypatch.setattr(istream, "fileno", fileno)
        monkeypatch.setattr(termios, "tcgetattr", tcgetattr)
        monkeypatch.setattr(termios, "tcsetattr", tcsetattr)

        io_mocker.expect_screen(
            [
                "> Enter password:   ",
                "                    ",
                "                    ",
                "                    ",
                "                    ",
            ],
        )
        io_mocker.expect_eq(lambda: mode[3], 0)
        io_mocker.expect_istream_readline("123 456\n", echo=False)

        with io_mocker.mock(), pytest.raises(termios.error, match=r"^error$"):
            assert (
                yuio.io.ask[yuio.parse.SecretString]("Enter password").data == "123 456"
            )

        assert mode == orig_mode

    @pytest.mark.windows
    def test_secret_windows(self, io_mocker: IOMocker, monkeypatch: pytest.MonkeyPatch):
        import msvcrt

        input_str = "X\b\b\basX\bd\0\0dsa\r\n"
        output_str = ""
        input_index = 0

        def getwch():
            nonlocal input_index
            assert input_index < len(input_str)
            c = input_str[input_index]
            input_index += 1
            return c

        def putwch(c):
            nonlocal output_str
            output_str += c

        monkeypatch.setattr(msvcrt, "getwch", getwch)
        monkeypatch.setattr(msvcrt, "putwch", putwch)

        io_mocker.expect_screen(
            [
                "> Enter password:   ",
                "                    ",
                "                    ",
                "                    ",
                "                    ",
            ],
        )

        with io_mocker.mock():
            assert (
                yuio.io.ask[yuio.parse.SecretString]("Enter password").data == "asddsa"
            )

        assert output_str == "*\b \b***\b \b****\r\n"

    @pytest.mark.windows
    def test_secret_windows_keyboard_interrupt(
        self, io_mocker: IOMocker, monkeypatch: pytest.MonkeyPatch
    ):
        import msvcrt

        input_str = "abc\003"
        output_str = ""
        input_index = 0

        def getwch():
            nonlocal input_index
            assert input_index < len(input_str)
            c = input_str[input_index]
            input_index += 1
            return c

        def putwch(c):
            nonlocal output_str
            output_str += c

        monkeypatch.setattr(msvcrt, "getwch", getwch)
        monkeypatch.setattr(msvcrt, "putwch", putwch)

        io_mocker.expect_screen(
            [
                "> Enter password:   ",
                "                    ",
                "                    ",
                "                    ",
                "                    ",
            ],
        )

        with io_mocker.mock(), pytest.raises(KeyboardInterrupt):
            assert (
                yuio.io.ask[yuio.parse.SecretString]("Enter password").data == "asddsa"
            )

        assert output_str == "***"


class TestAskNonInteractive:
    @pytest.fixture
    def term(self, ostream: io.StringIO, istream: _t.TextIO) -> yuio.term.Term:
        istream.readable = lambda: False
        return yuio.term.Term(
            ostream,
            istream,
            color_support=yuio.term.ColorSupport.ANSI_TRUE,
            ostream_is_tty=False,
            istream_is_tty=False,
            is_unicode=True,
        )

    def test_default(self, io_mocker: IOMocker):
        with io_mocker.mock():
            assert yuio.io.ask("Meow?", default="Meow!") == "Meow!"

    def test_default_non_interactive(self, io_mocker: IOMocker):
        with io_mocker.mock():
            assert (
                yuio.io.ask("Meow?", default="Meow!", default_non_interactive="Boop!")
                == "Boop!"
            )

    def test_no_default(self, io_mocker: IOMocker):
        with io_mocker.mock():
            with pytest.raises(
                yuio.io.UserIoError, match=r"non-interactive environment"
            ):
                yuio.io.ask("Meow?")


class TestWaitForUser:
    def test_simple(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Press enter to      ",
                "continue            ",
            ]
        )
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            yuio.io.wait_for_user()

    def test_msg(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Slam that Enter     ",
                "button!             ",
            ]
        )
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            yuio.io.wait_for_user("Slam that Enter button!")

    def test_msg_format(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Slam that Enter     ",
                "button!             ",
            ]
        )
        io_mocker.key(yuio.widget.Key.ENTER)

        with io_mocker.mock():
            yuio.io.wait_for_user("Slam that %s button!", "Enter")


class TestWaitForUserNoColor:
    @pytest.fixture
    def term(self, ostream: io.StringIO, istream: _t.TextIO) -> yuio.term.Term:
        return yuio.term.Term(
            ostream,
            istream,
            color_support=yuio.term.ColorSupport.NONE,
            ostream_is_tty=True,
            istream_is_tty=True,
            is_unicode=True,
        )

    def test_simple(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "> Press enter to    ",
                "  continue          ",
            ]
        )
        io_mocker.expect_istream_readline("\n")

        with io_mocker.mock():
            yuio.io.wait_for_user()

    def test_add_space(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "> Press enter to    ",
                "  continue          ",
            ]
        )
        io_mocker.expect_istream_readline(".\n")
        io_mocker.expect_screen(
            [
                "> Press enter to    ",
                "  continue .        ",
            ]
        )

        with io_mocker.mock():
            yuio.io.wait_for_user()

    def test_strip_end(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "> Ends with spaces  ",
            ]
        )
        io_mocker.expect_istream_readline(".\n")
        io_mocker.expect_screen(
            [
                "> Ends with spaces .",
            ]
        )

        with io_mocker.mock():
            yuio.io.wait_for_user("Ends with spaces   ")

    def test_msg(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "> Slam that Enter   ",
                "  button!           ",
            ]
        )
        io_mocker.expect_istream_readline("\n")

        with io_mocker.mock():
            yuio.io.wait_for_user("Slam that Enter button!")

    def test_msg_format(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "> Slam that Enter   ",
                "  button!           ",
            ]
        )
        io_mocker.expect_istream_readline("\n")

        with io_mocker.mock():
            yuio.io.wait_for_user("Slam that %s button!", "Enter")


class TestWaitForUserNonInteractive:
    @pytest.fixture
    def term(self, ostream: io.StringIO, istream: _t.TextIO) -> yuio.term.Term:
        istream.readable = lambda: False
        return yuio.term.Term(
            ostream,
            istream,
            color_support=yuio.term.ColorSupport.ANSI_TRUE,
            ostream_is_tty=False,
            istream_is_tty=False,
            is_unicode=True,
        )

    def test_wait(self, io_mocker: IOMocker):
        io_mocker.expect_screen([])
        with io_mocker.mock():
            yuio.io.wait_for_user()


class TestDetectEditor:
    @pytest.fixture(autouse=True)
    def setup_env(self, monkeypatch):
        monkeypatch.delenv("EDITOR", raising=False)
        monkeypatch.delenv("VISUAL", raising=False)
        return

    @pytest.mark.linux
    @pytest.mark.darwin
    def test_env_editor(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setenv("EDITOR", "foobar")
        assert yuio.io.detect_editor() == "foobar"

    @pytest.mark.linux
    @pytest.mark.darwin
    def test_env_visual(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setenv("EDITOR", "foobar")
        monkeypatch.setenv("VISUAL", "visual")
        assert yuio.io.detect_editor() == "visual"

    @pytest.mark.windows
    def test_env_windows(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setenv("EDITOR", "foobar")
        monkeypatch.setenv("VISUAL", "visual")
        monkeypatch.setattr(
            "shutil.which", lambda exc: exc if exc == "notepad" else None
        )
        assert yuio.io.detect_editor() == "notepad"

    @pytest.mark.parametrize(
        "editor",
        [
            "vi",
            "nano",
            "notepad",
        ],
    )
    def test_which(self, editor: str, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr("shutil.which", lambda exc: exc if exc == editor else None)
        assert yuio.io.detect_editor() == editor

    def test_fail(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr("shutil.which", lambda exc: None)
        assert yuio.io.detect_editor() is None

    def test_fallbacks(self, monkeypatch: pytest.MonkeyPatch):
        seen = []
        monkeypatch.setattr("shutil.which", lambda exc: (seen.append(exc), None)[1])
        assert yuio.io.detect_editor(fallbacks=["fallback 1", "fallback 2"]) is None
        assert seen == ["fallback 1", "fallback 2"]


@pytest.mark.linux
@pytest.mark.darwin
class TestEditUnix:
    _EDITOR = "echo ' edited' >>"

    def test_simple(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr("yuio.io.detect_editor", lambda _: self._EDITOR)
        assert yuio.io.edit("foobar") == "foobar edited\n"

    def test_editor(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr("yuio.io.detect_editor", lambda _: None)
        assert yuio.io.edit("foobar", editor=self._EDITOR) == "foobar edited\n"

    def test_editor_error(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr("yuio.io.detect_editor", lambda _: "exit 1; cat")
        with pytest.raises(yuio.io.UserIoError, match=r"returned exit code 1"):
            assert yuio.io.edit("foobar")

    def test_editor_signal(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr("yuio.io.detect_editor", lambda _: "kill -9 $$; cat")
        with pytest.raises(yuio.io.UserIoError, match=r"died with SIGKILL"):
            assert yuio.io.edit("foobar")

    def test_no_editor(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr("yuio.io.detect_editor", lambda _: None)
        with pytest.raises(yuio.io.UserIoError, match=r"Can't find a usable editor"):
            assert yuio.io.edit("foobar")

    def test_comments(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr("yuio.io.detect_editor", lambda _: self._EDITOR)
        assert (
            yuio.io.edit("# foo\n  # bar\nbaz #", comment_marker="#")
            == "baz # edited\n"
        )
        assert yuio.io.edit("foo\n#", comment_marker="#") == "foo\n"

    def test_comments_long_marker(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr("yuio.io.detect_editor", lambda _: self._EDITOR)
        assert (
            yuio.io.edit("// foo\n  # bar\nbaz //", comment_marker="//")
            == "  # bar\nbaz // edited\n"
        )
        assert yuio.io.edit("foo\n//", comment_marker="//") == "foo\n"

    def test_comments_long_marker_special_symbols(
        self, monkeypatch: pytest.MonkeyPatch
    ):
        monkeypatch.setattr("yuio.io.detect_editor", lambda _: self._EDITOR)
        assert yuio.io.edit("a\nb\n[ab]", comment_marker="[ab]") == "a\nb\n"

    def test_file_removed(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr("yuio.io.detect_editor", lambda _: "rm")
        with pytest.raises(yuio.io.UserIoError, match=r"can't read back edited file"):
            yuio.io.edit("foo")


@pytest.mark.linux
class TestEditLinux:
    def test_editor_unknown_signal(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr("yuio.io.detect_editor", lambda _: "kill -60 $$; cat")
        with pytest.raises(yuio.io.UserIoError, match=r"died with unknown signal -60"):
            assert yuio.io.edit("foobar")


@pytest.mark.windows
class TestEditWindows:
    SCRIPT = textwrap.dedent(
        """
        @echo off
        echo Edited content> %1
        """
    )

    def test_edit(self, tmp_path, monkeypatch):
        script = tmp_path / "edit.bat"
        script.write_text(self.SCRIPT)

        monkeypatch.setattr("yuio.io.detect_editor", lambda _: str(script))
        assert yuio.io.edit("foo").strip() == "Edited content"

    def test_no_editor(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr("yuio.io.detect_editor", lambda _: None)
        with pytest.raises(yuio.io.UserIoError, match=r"Can't find a usable editor"):
            assert yuio.io.edit("foobar")

    def test_editor_not_found(self):
        with pytest.raises(yuio.io.UserIoError, match=r"no such file or directory"):
            yuio.io.edit("foo", editor="/foo/bar")


class TestTask:
    @pytest.fixture
    def width(self):
        return 40

    def test_simple(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "⣿ task                                  ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "> task - done                           ",
            ]
        )

        with io_mocker.mock():
            with yuio.io.Task("task"):
                io_mocker.mark()

    def test_format(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "⣿ task 1                                ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "> task 1 - done                         ",
            ]
        )

        with io_mocker.mock():
            with yuio.io.Task("task `%s`", 1):
                io_mocker.mark()

    def test_print_while_in_task(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "⣿ task 1                                ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "hello world!                            ",
                "⣿ task 1                                ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "hello world!                            ",
                "> task 1 - done                         ",
                "hello world 2!                          ",
            ]
        )

        with io_mocker.mock():
            with yuio.io.Task("task `%s`", 1):
                io_mocker.mark()
                yuio.io.info("hello world!")
                io_mocker.mark()
            yuio.io.info("hello world 2!")

    def test_error_in_task(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "⣿ task                                  ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "> task - error                          ",
            ]
        )

        with io_mocker.mock():
            with pytest.raises(RuntimeError, match=r"eh..."):
                with yuio.io.Task("task"):
                    io_mocker.mark()
                    raise RuntimeError("eh...")

    def test_manual_error(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "⣿ task                                  ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "> task - error                          ",
            ]
        )

        with io_mocker.mock():
            task = yuio.io.Task("task")
            io_mocker.mark()
            task.error()

    def test_manual_done(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "⣿ task                                  ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "> task - done                           ",
            ]
        )

        with io_mocker.mock():
            task = yuio.io.Task("task")
            io_mocker.mark()
            task.done()

    def test_comment(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "⣿ task                                  ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "⣿ task - yaaay!                         ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "> task - done                           ",
            ]
        )

        with io_mocker.mock():
            with yuio.io.Task("task") as task:
                io_mocker.mark()
                task.comment("yaaay!")
                io_mocker.mark()

    def test_comment_format(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "⣿ task                                  ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "⣿ task - yaaay 1!                       ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "> task - done                           ",
            ]
        )

        with io_mocker.mock():
            with yuio.io.Task("task") as task:
                io_mocker.mark()
                task.comment("yaaay `%s`!", 1)
                io_mocker.mark()

    def test_comment_format_chached(self, io_mocker: IOMocker):
        class Fmt:
            i = 0

            def __str__(self):
                self.i += 1
                return str(self.i)

        fmt = Fmt()

        io_mocker.expect_screen(
            [
                "⣿ task                                  ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "⣿ task - yaaay 1!                       ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "■■□□□ task - 50.00% - yaaay 1!          ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "■■□□□ task - 50.00% - yaaay 2!          ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "> task - done                           ",
            ]
        )

        with io_mocker.mock():
            with yuio.io.Task("task") as task:
                io_mocker.mark()
                task.comment("yaaay `%s`!", fmt)
                io_mocker.mark()
                task.progress(0.5)
                io_mocker.mark()
                task.comment("yaaay `%s`!", fmt)
                io_mocker.mark()

    @pytest.mark.parametrize(
        ("args", "kwargs", "comment", "expected"),
        [
            (
                (0.3,),
                {},
                None,
                [
                    "■■□□□ task - 30.00%                     ",
                ],
            ),
            (
                (0.3,),
                {"unit": "whatever"},
                None,
                [
                    "■■□□□ task - 30.00%                     ",
                ],
            ),
            (
                (0.3,),
                {"ndigits": 0},
                None,
                [
                    "■■□□□ task - 30%                        ",
                ],
            ),
            (
                (11, 15),
                {},
                None,
                [
                    "■■■■□ task - 11/15                      ",
                ],
            ),
            (
                (10.5, 15),
                {},
                None,
                [
                    "■■■■□ task - 10.50/15.00                ",
                ],
            ),
            (
                (11, 15.0),
                {},
                None,
                [
                    "■■■■□ task - 11.00/15.00                ",
                ],
            ),
            (
                (10.5, 15),
                {"unit": "Kg"},
                None,
                [
                    "■■■■□ task - 10.50/15.00Kg              ",
                ],
            ),
            (
                (10.5, 15),
                {"ndigits": 3},
                None,
                [
                    "■■■■□ task - 10.500/15.000              ",
                ],
            ),
            (
                (10.5, 15),
                {"ndigits": 3, "unit": "Kg"},
                None,
                [
                    "■■■■□ task - 10.500/15.000Kg            ",
                ],
            ),
            (
                (0.3,),
                {},
                "comment!",
                [
                    "■■□□□ task - 30.00% - comment!          ",
                ],
            ),
            (
                (0.3,),
                {"unit": "whatever"},
                "comment!",
                [
                    "■■□□□ task - 30.00% - comment!          ",
                ],
            ),
            (
                (0.3,),
                {"ndigits": 0},
                "comment!",
                [
                    "■■□□□ task - 30% - comment!             ",
                ],
            ),
            (
                (11, 15),
                {},
                "comment!",
                [
                    "■■■■□ task - 11/15 - comment!           ",
                ],
            ),
            (
                (10.5, 15),
                {},
                "comment!",
                [
                    "■■■■□ task - 10.50/15.00 - comment!     ",
                ],
            ),
            (
                (10.5, 15),
                {"unit": "Kg"},
                "comment!",
                [
                    "■■■■□ task - 10.50/15.00Kg - comment!   ",
                ],
            ),
            (
                (10.5, 15),
                {"ndigits": 3},
                "comment!",
                [
                    "■■■■□ task - 10.500/15.000 - comment!   ",
                ],
            ),
            (
                (10.5, 15),
                {"ndigits": 3, "unit": "Kg"},
                "comment!",
                [
                    "■■■■□ task - 10.500/15.000Kg - comment! ",
                ],
            ),
        ],
    )
    def test_progress(self, io_mocker: IOMocker, args, kwargs, comment, expected):
        io_mocker.expect_screen(
            [
                "⣿ task                                  ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(expected)
        io_mocker.expect_mark()

        with io_mocker.mock():
            with yuio.io.Task("task") as task:
                io_mocker.mark()
                task.progress(*args, **kwargs)
                if comment:
                    task.comment(comment)
                io_mocker.mark()

    @pytest.mark.parametrize(
        ("args", "kwargs", "expected"),
        [
            (
                (50, 100),
                {},
                [
                    "■■□□□ task - 50.00/100.00B              ",
                ],
            ),
            (
                (0.50, 1),
                {},
                [
                    "■■□□□ task - 0.50/1.00B                 ",
                ],
            ),
            (
                (50, 100 * 1024),
                {},
                [
                    "□□□□□ task - 50.00B/100.00K             ",
                ],
            ),
            (
                (50, 100 * 1024 + 100),
                {},
                [
                    "□□□□□ task - 50.00B/100.10K             ",
                ],
            ),
            (
                (50 * 1024, 100 * 1024),
                {},
                [
                    "■■□□□ task - 50.00/100.00K              ",
                ],
            ),
            (
                (50 * 1024, 100.1 * 1024),
                {},
                [
                    "■■□□□ task - 50.00/100.10K              ",
                ],
            ),
            (
                (50, 100.501 * 1024 * 1024),
                {},
                [
                    "□□□□□ task - 50.00B/100.50M             ",
                ],
            ),
            (
                (50, 100.501 * 1024 * 1024),
                {"ndigits": 4},
                [
                    "□□□□□ task - 50.0000B/100.5010M         ",
                ],
            ),
            (
                (1 * 1024 * 1024 * 1024, 100.5 * 1024 * 1024 * 1024),
                {},
                [
                    "□□□□□ task - 1.00/100.50G               ",
                ],
            ),
            (
                (1 * 1024 * 1024 * 1024, 5000 * 1024 * 1024 * 1024 * 1024 * 1024),
                {},
                [
                    "□□□□□ task - 1.00G/5000.00P             ",
                ],
            ),
        ],
    )
    def test_progress_size(self, io_mocker: IOMocker, args, kwargs, expected):
        io_mocker.expect_screen(
            [
                "⣿ task                                  ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(expected)
        io_mocker.expect_mark()

        with io_mocker.mock():
            with yuio.io.Task("task") as task:
                io_mocker.mark()
                task.progress_size(*args, **kwargs)
                io_mocker.mark()

    @pytest.mark.parametrize(
        ("args", "kwargs", "expected"),
        [
            (
                (50, 100),
                {},
                [
                    "■■□□□ task - 50.00/100.00               ",
                ],
            ),
            (
                (50000, 100000),
                {},
                [
                    "■■□□□ task - 50.00K/100.00K             ",
                ],
            ),
            (
                (0.01052, 0.1),
                {},
                [
                    "■□□□□ task - 10.52m/100.00m             ",
                ],
            ),
            (
                (0.000000000000000000000001, 1000000000000000000000000),
                {},
                [
                    "□□□□□ task - 1.00y/1.00Y                ",
                ],
            ),
            (
                (0.0000000000000000000000001, 1000200000000000000000000000),
                {},
                [
                    "□□□□□ task - 0.10y/1000.20Y             ",
                ],
            ),
            (
                (50, 100),
                {"unit": "V"},
                [
                    "■■□□□ task - 50.00V/100.00V             ",
                ],
            ),
            (
                (0.00050, 0.001001),
                {"unit": "V"},
                [
                    "■■□□□ task - 500.00uV/1.00mV            ",
                ],
            ),
            (
                (0.00050, 0.001001),
                {"ndigits": 4},
                [
                    "■■□□□ task - 500.0000u/1.0010m          ",
                ],
            ),
        ],
    )
    def test_progress_scale(self, io_mocker: IOMocker, args, kwargs, expected):
        io_mocker.expect_screen(
            [
                "⣿ task                                  ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(expected)
        io_mocker.expect_mark()

        with io_mocker.mock():
            with yuio.io.Task("task") as task:
                io_mocker.mark()
                task.progress_scale(*args, **kwargs)
                io_mocker.mark()

    def test_reset_progress(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "⣿ task                                  ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "■■□□□ task - 2/5                        ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "⣿ task                                  ",
            ]
        )
        io_mocker.expect_mark()

        with io_mocker.mock():
            with yuio.io.Task("task") as task:
                io_mocker.mark()
                task.progress(2, 5)
                io_mocker.mark()
                task.progress(None)
                io_mocker.mark()

    @pytest.mark.parametrize(
        ("kwargs", "expected_format"),
        [
            ({}, "{}/{}"),
            ({"ndigits": 2}, "{}.00/{}.00"),
            ({"unit": "x"}, "{}/{}x"),
            ({"unit": "y", "ndigits": 1}, "{}.0/{}.0y"),
        ],
    )
    def test_iter(self, io_mocker: IOMocker, kwargs, expected_format: str):
        io_mocker.expect_screen(
            [
                "⣿ task                                  ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                f"□□□□□ task - {expected_format.format(0, 5):<27}",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                f"■□□□□ task - {expected_format.format(1, 5):<27}",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                f"■■□□□ task - {expected_format.format(2, 5):<27}",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                f"■■■□□ task - {expected_format.format(3, 5):<27}",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                f"■■■■□ task - {expected_format.format(4, 5):<27}",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                f"■■■■■ task - {expected_format.format(5, 5):<27}",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "> task - done                           ",
            ]
        )

        elems = [1, 2, "A", "B", object()]

        with io_mocker.mock():
            with yuio.io.Task("task") as task:
                io_mocker.mark()
                for i, elem in enumerate(task.iter(elems, **kwargs)):
                    assert elem is elems[i]
                    io_mocker.mark()
                io_mocker.mark()

    def test_multiple_tasks(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "⣿ T1                                    ",
                "⣿ T2                                    ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "■■□□□ T1 - 2/5                          ",
                "■■■■□ T2 - 4/5                          ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "foo                                     ",
                "■■□□□ T1 - 2/5                          ",
                "■■■■□ T2 - 4/5                          ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "foo                                     ",
                "> T2 - done                             ",
                "■■□□□ T1 - 2/5                          ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "foo                                     ",
                "> T2 - done                             ",
                "bar                                     ",
                "■■□□□ T1 - 2/5                          ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "foo                                     ",
                "> T2 - done                             ",
                "bar                                     ",
                "> T1 - done                             ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "foo                                     ",
                "> T2 - done                             ",
                "bar                                     ",
                "> T1 - done                             ",
                "baz                                     ",
            ]
        )

        with io_mocker.mock():
            t1 = yuio.io.Task("T1")
            t2 = yuio.io.Task("T2")
            io_mocker.mark()
            t1.progress(2, 5)
            t2.progress(4, 5)
            io_mocker.mark()
            yuio.io.info("foo")
            io_mocker.mark()
            t2.done()
            io_mocker.mark()
            yuio.io.info("bar")
            io_mocker.mark()
            t1.done()
            io_mocker.mark()
            yuio.io.info("baz")

    def test_subtask(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "⣿ T1                                    ",
                "  ⣿ T1.1                                ",
                "  ⣿ T1.2                                ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "⣿ T1                                    ",
                "  ⣿ T1.1                                ",
                "  ⣿ T1.2 - done                         ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "foo                                     ",
                "⣿ T1                                    ",
                "  ⣿ T1.1                                ",
                "  ⣿ T1.2 - done                         ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "foo                                     ",
                "⣿ T1                                    ",
                "  ■■□□□ T1.1 - 2/5                      ",
                "  ⣿ T1.2 - done                         ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "foo                                     ",
                "> T1 - done                             ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "foo                                     ",
                "> T1 - done                             ",
                "bar                                     ",
            ]
        )

        with io_mocker.mock():
            t1 = yuio.io.Task("T1")
            t11 = t1.subtask("T1.1")
            t12 = t1.subtask("T1.2")
            io_mocker.mark()
            t12.done()
            io_mocker.mark()
            yuio.io.info("foo")
            io_mocker.mark()
            t11.progress(2, 5)
            io_mocker.mark()
            t1.done()
            io_mocker.mark()
            yuio.io.info("bar")

    def test_progressbar_decoration(self, theme: yuio.theme.Theme, io_mocker: IOMocker):
        theme.set_msg_decoration_unicode("progress_bar/start_symbol", "[")
        theme.set_msg_decoration_unicode("progress_bar/end_symbol", "]")
        theme.set_msg_decoration_unicode("progress_bar/done_symbol", ">")
        theme.set_msg_decoration_unicode("progress_bar/pending_symbol", ".")
        theme.set_msg_decoration_unicode("progress_bar/transition_pattern", "")
        theme.progress_bar_width = 17

        io_mocker.expect_screen(
            [
                "[>>>>>>.........] Task - 40.00%         ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "[>>>>>>.........] Task - 41.00%         ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "[>>>>>>>........] Task - 45.00%         ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "[>>>>>>>>.......] Task - 50.00%         ",
            ]
        )
        io_mocker.expect_mark()

        with io_mocker.mock():
            with yuio.io.Task("Task") as task:
                task.progress(0.40)
                io_mocker.mark()
                task.progress(0.41)
                io_mocker.mark()
                task.progress(0.45)
                io_mocker.mark()
                task.progress(0.50)
                io_mocker.mark()

    def test_progressbar_decoration_transition(
        self, theme: yuio.theme.Theme, io_mocker: IOMocker
    ):
        theme.set_msg_decoration_unicode("progress_bar/start_symbol", "[")
        theme.set_msg_decoration_unicode("progress_bar/end_symbol", "]")
        theme.set_msg_decoration_unicode("progress_bar/done_symbol", ">")
        theme.set_msg_decoration_unicode("progress_bar/pending_symbol", ".")
        theme.set_msg_decoration_unicode(
            "progress_bar/transition_pattern", "9876543210"
        )
        theme.progress_bar_width = 17

        io_mocker.expect_screen(
            [
                "[>>>>>>0........] Task - 40.00%         ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "[>>>>>>2........] Task - 41.00%         ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "[>>>>>>8........] Task - 45.00%         ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "[>>>>>>>5.......] Task - 50.00%         ",
            ]
        )
        io_mocker.expect_mark()

        with io_mocker.mock():
            with yuio.io.Task("Task") as task:
                task.progress(0.40)
                io_mocker.mark()
                task.progress(0.41)
                io_mocker.mark()
                task.progress(0.45)
                io_mocker.mark()
                task.progress(0.50)
                io_mocker.mark()


class TestTaskNonInteractive:
    pass


class TestTaskBackground:
    is_foreground = False

    @pytest.fixture
    def width(self):
        return 40

    @pytest.fixture(autouse=True)
    def setup_is_foreground(self, monkeypatch: pytest.MonkeyPatch):
        def is_foreground(*_, **__):
            return self.is_foreground

        monkeypatch.setattr("yuio.term._is_foreground", is_foreground)

    def test_background(self, io_mocker):
        io_mocker.expect_screen(
            [
                "> task - running                        ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "> task - running                        ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "> task - running                        ",
                "> task - done                           ",
            ]
        )

        with io_mocker.mock():
            with yuio.io.Task("task") as task:
                io_mocker.mark()
                task.progress(0.5)
                io_mocker.mark()

    def test_background_switch(self, io_mocker):
        io_mocker.expect_screen(
            [
                "> task - running                        ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "> task - running                        ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "> task - running                        ",
                "■■□□□ task - 50.00%                     ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "> task - running                        ",
                "■■□□□ task - 50.00%                     ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "> task - running                        ",
                "■■□□□ task - 50.00%                     ",
                "> task - done                           ",
            ]
        )

        with io_mocker.mock():
            with yuio.io.Task("task") as task:
                io_mocker.mark()
                task.progress(0.5)
                io_mocker.mark()
                self.is_foreground = True
                task.progress(0.5)
                io_mocker.mark()
                self.is_foreground = False
                task.progress(0.5)
                io_mocker.mark()

    def test_background_switch_with_subtasks(self, io_mocker):
        io_mocker.expect_screen(
            [
                "> task - running                        ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "> task - running                        ",
                "> a - running                           ",
                "> b - running                           ",
                "> c - running                           ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "> task - running                        ",
                "> a - running                           ",
                "> b - running                           ",
                "> c - running                           ",
                "⣿ task                                  ",
                "  ⣿ a - comment                         ",
                "  ⣿ b                                   ",
                "  ⣿ c                                   ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "> task - running                        ",
                "> a - running                           ",
                "> b - running                           ",
                "> c - running                           ",
                "⣿ task                                  ",
                "  ⣿ a - done                            ",
                "  ⣿ b                                   ",
                "  ⣿ c                                   ",
                "> b - done                              ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "> task - running                        ",
                "> a - running                           ",
                "> b - running                           ",
                "> c - running                           ",
                "⣿ task                                  ",
                "  ⣿ a - done                            ",
                "  ⣿ b                                   ",
                "  ⣿ c                                   ",
                "> b - done                              ",
                "⣿ task                                  ",
                "  ⣿ a - done                            ",
                "  ⣿ b - done                            ",
                "  ⣿ c - done                            ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "> task - running                        ",
                "> a - running                           ",
                "> b - running                           ",
                "> c - running                           ",
                "⣿ task                                  ",
                "  ⣿ a - done                            ",
                "  ⣿ b                                   ",
                "  ⣿ c                                   ",
                "> b - done                              ",
                "⣿ task                                  ",
                "  ⣿ a - done                            ",
                "  ⣿ b - done                            ",
                "  ⣿ c - done                            ",
                "> task - done                           ",
            ]
        )
        with io_mocker.mock():
            with yuio.io.Task("task") as task:
                io_mocker.mark()
                a = task.subtask("a")
                b = task.subtask("b")
                c = task.subtask("c")
                io_mocker.mark()
                self.is_foreground = True
                a.comment("comment")
                io_mocker.mark()
                a.done()
                self.is_foreground = False
                b.done()
                io_mocker.mark()
                self.is_foreground = True
                c.done()
                io_mocker.mark()
                self.is_foreground = False


class TestSuspendOutput:
    @pytest.mark.parametrize(
        ("meth", "args", "expected"),
        [
            (
                "info",
                ("Bar.",),
                [
                    "Bar.                ",
                ],
            ),
            (
                "warning",
                ("Bar.",),
                [
                    "Bar.                ",
                ],
            ),
            (
                "success",
                ("Bar.",),
                [
                    "Bar.                ",
                ],
            ),
            (
                "error",
                ("Bar.",),
                [
                    "Bar.                ",
                ],
            ),
            (
                "error_with_tb",
                ("Bar.",),
                [
                    "Bar.                ",
                ],
            ),
            (
                "heading",
                ("Bar.",),
                [
                    "                    ",
                    "⣿ Bar.              ",
                    "                    ",
                ],
            ),
            (
                "md",
                ("Bar.",),
                [
                    "Bar.                ",
                ],
            ),
            (
                "br",
                (),
                [
                    "                    ",
                ],
            ),
            (
                "raw",
                (yuio.string.ColorizedString("Bar."),),
                [
                    "Bar.                ",
                ],
            ),
        ],
    )
    def test_simple(self, io_mocker: IOMocker, meth, args, expected):
        io_mocker.expect_screen(
            [
                "Foo.                ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "Foo.                ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "Foo.                ",
                *expected,
            ]
        )

        with io_mocker.mock():
            yuio.io.info("Foo.")
            with yuio.io.SuspendOutput():
                io_mocker.mark()
                getattr(yuio.io, meth)(*args)
                io_mocker.mark()

    @pytest.mark.parametrize(
        "meth",
        [
            lambda s: sys.stderr.write(s + "\n"),
            lambda s: sys.stdout.write(s + "\n"),
            lambda s: print(s),
        ],
    )
    def test_streams(self, io_mocker: IOMocker, meth):
        io_mocker.expect_screen(
            [
                "Foo.                ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "Foo.                ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "Foo.                ",
                "Bar.                ",
            ]
        )

        with io_mocker.mock(wrap_streams=True):
            yuio.io.info("Foo.")
            with yuio.io.SuspendOutput():
                io_mocker.mark()
                meth("Bar.")
                io_mocker.mark()

    @pytest.mark.parametrize(
        ("meth", "args", "expected"),
        [
            (
                "warning",
                ("Bar.",),
                [
                    "Bar.                ",
                ],
            ),
            (
                "success",
                ("Bar.",),
                [
                    "Bar.                ",
                ],
            ),
            (
                "error",
                ("Bar.",),
                [
                    "Bar.                ",
                ],
            ),
            (
                "error_with_tb",
                ("Bar.",),
                [
                    "Bar.                ",
                ],
            ),
            (
                "md",
                ("Bar.",),
                [
                    "Bar.                ",
                ],
            ),
            (
                "br",
                (),
                [
                    "                    ",
                ],
            ),
            (
                "raw",
                (yuio.string.ColorizedString("Bar.\n"),),
                [
                    "Bar.                ",
                ],
            ),
        ],
    )
    def test_ignore(self, io_mocker: IOMocker, meth, args, expected):
        io_mocker.expect_screen(
            [
                "Foo.                ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "Foo.                ",
                *expected,
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "Foo.                ",
                *expected,
                *expected,
            ]
        )

        with io_mocker.mock():
            yuio.io.info("Foo.")
            with yuio.io.SuspendOutput() as o:
                io_mocker.mark()
                getattr(yuio.io, meth)(*args)
                getattr(o, meth)(*args)
                io_mocker.mark()

    def test_task(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Foo.                ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "Foo.                ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "Foo.                ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "Foo.                ",
                "> Task - done       ",
            ]
        )

        with io_mocker.mock():
            yuio.io.info("Foo.")
            io_mocker.mark()
            with yuio.io.SuspendOutput():
                task = yuio.io.Task("Task")
                io_mocker.mark()
                task.done()
                io_mocker.mark()

    def test_task_start_before_suspended(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Foo.                ",
                "⣿ Task              ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "Foo.                ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "Foo.                ",
                "⣿ Task              ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "Foo.                ",
                "> Task - done       ",
            ]
        )

        with io_mocker.mock():
            yuio.io.info("Foo.")
            task = yuio.io.Task("Task")
            io_mocker.mark()
            with yuio.io.SuspendOutput():
                io_mocker.mark()
            io_mocker.mark()
            task.done()

    def test_task_finish_while_suspended(self, io_mocker: IOMocker):
        io_mocker.expect_screen(
            [
                "Foo.                ",
                "⣿ Task              ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "Foo.                ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "Foo.                ",
            ]
        )
        io_mocker.expect_mark()
        io_mocker.expect_screen(
            [
                "Foo.                ",
                "> Task - done       ",
            ]
        )

        with io_mocker.mock():
            yuio.io.info("Foo.")
            task = yuio.io.Task("Task")
            io_mocker.mark()
            with yuio.io.SuspendOutput():
                io_mocker.mark()
                task.done()
                io_mocker.mark()


class TestLog:
    @pytest.fixture(autouse=True)
    def setup_logger(self):
        import logging

        self.logger = logging.getLogger("yuio.test.test_log")
        self.logger.setLevel(logging.DEBUG)
        handler = yuio.io.Handler()
        self.logger.addHandler(handler)

        yield

        self.logger.removeHandler(handler)

    def test_simple(self, ostream: io.StringIO):
        self.logger.debug("debug")
        self.logger.info("info")
        self.logger.warning("warning")
        self.logger.error("error")
        self.logger.critical("critical")

        value = ostream.getvalue()
        assert "yuio.test.test_log DEBUG debug" in value
        assert "yuio.test.test_log INFO info" in value
        assert "yuio.test.test_log WARNING warning" in value
        assert "yuio.test.test_log ERROR error" in value
        assert "yuio.test.test_log CRITICAL critical" in value

    def test_tb(self, ostream: io.StringIO):
        self.logger.debug("debug")
        try:
            raise RuntimeError("oh no!")
        except RuntimeError:
            self.logger.exception("something went wrong")

        value = ostream.getvalue()
        assert "something went wrong" in value
        assert "oh no!" in value

    def test_stack(self, ostream: io.StringIO):
        self.logger.info("hi there!", stack_info=True)

        value = ostream.getvalue()
        assert "hi there!" in value
        assert 'self.logger.info("hi there!", stack_info=True)' in value

    def test_colors(self, theme: yuio.theme.Theme, io_mocker: IOMocker):
        self.logger.handlers[0].setFormatter(
            yuio.io.Formatter(
                "%(name)s %(levelname)s %(add_data)d %(add_data_2)s %(colMessage)s",
                defaults={
                    "add_data": 10,
                    "add_data_2": "asd",
                },
            )
        )
        theme.set_color("log", "yellow")
        theme.set_color("log:info", "cyan")
        theme.set_color("log/add_data:info", "red")
        theme.set_color("log/name:info", "green")
        theme.set_color("log/levelname:info", "blue")
        theme.set_color("log/colMessage:info", "magenta")

        io_mocker.expect_screen(
            [
                "yuio.test.test_log I",
                "NFO 10 asd msg arg  ",
                "yuio.test.test_log E",
                "RROR 10 asd m2 `arg`",
            ],
            [
                "ggggggggggggggggggcb",
                "bbbcrrcccccmmmmmmm  ",
                "yyyyyyyyyyyyyyyyyyyy",
                "yyyyyyyyyyyyyyyyYYYy",
            ],
        )

        with io_mocker.mock():
            self.logger.info("msg %s", "arg")
            self.logger.error(
                "m2 `%s`", yuio.io.WithBaseColor("arg", base_color="bold")
            )



import shlex
import subprocess

from .conftest import extract_results

import pytest

SCRIPT = """
COMP_CWORD={cword}
COMP_POINT={prefix_len}
COMP_WORDS=(comptest {args})
COMP_LINE="${{COMP_WORDS[*]}}"
. ./test.bash
"""


@pytest.mark.full
@pytest.mark.linux
def test_bash(test_cases, data_regression):
    subprocess.check_call(["comptest", "--completions", "bash"])

    results = []
    for cword, prefix_len, _, args, orig in test_cases:
        print(args)
        try:
            result = subprocess.check_output(
                [
                    "bash",
                    "-c",
                    SCRIPT.format(
                        cword=cword + 1,
                        prefix_len=prefix_len + 9,
                        args=" ".join(map(shlex.quote, args)),
                    ),
                ],
            ).decode()
        except subprocess.CalledProcessError as e:
            results.append(dict(cmd=orig, err=str(e)))
        else:
            results.append(dict(cmd=orig, results=extract_results(result)))
    data_regression.check(results)



import datetime
import os
import pathlib
import sys

os.environ["__YUIO_SPHINX_BUILD"] = "1"
sys.path.append(str(pathlib.Path(__file__).parent / "_code"))
sys.path.append(str(pathlib.Path(__file__).parent / "_extensions"))

import yuio

# -- Project information -----------------------------------------------------

project = "Yuio"
copyright = f"{datetime.date.today().year}, Tamika Nomara"
author = "Tamika Nomara"
release = version = yuio.__version__


# -- General configuration ---------------------------------------------------

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.githubpages",
    "sphinx.ext.todo",
    "yuio.ext.sphinx",
    "sphinx_design",
    "sphinx_vhs",
    "sphinx_syntax",
    "nice_toc",
    "api_ref",
    "yuio_objects",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master/", None),
}
nitpick_ignore_regex = [
    (
        r"py:obj",
        r"(.*\.)?([A-Z]\w?|[A-Z]+_co|Cmp|SupportsLt|TAst|NamespaceT|ConfigT|_[^.]*)",
    ),
    (
        r"py:class",
        r"(.*\.)?([A-Z]\w?|[A-Z]+_co|Cmp|SupportsLt|TAst|NamespaceT|ConfigT|_[^.]*)",
    ),
]
autodoc_typehints_format = "short"
autodoc_member_order = "bysource"
autodoc_inherit_docstrings = False
# autodoc_typehints = "description"  # maybe sphinx 9 fixes this mess?
autodoc_type_aliases = {
    "yuio.Disabled": "~yuio.DISABLED",
    "DISABLED": "~yuio.DISABLED",
    "yuio.Missing": "~yuio.MISSING",
    "MISSING": "~yuio.MISSING",
    "yuio.Positional": "~yuio.POSITIONAL",
    "POSITIONAL": "~yuio.POSITIONAL",
    "yuio.Group": "~yuio.GROUP",
    "GROUP": "~yuio.GROUP",
    **{
        name: path
        for path in [
            "~yuio.app.OptionCtor",
            "~yuio.io.ExcInfo",
            "~yuio.widget.ActionKey",
            "~yuio.widget.ActionKeys",
            "~yuio.widget.Action",
            "~yuio.string.Colorable",
            "~yuio.string.RawString",
            "~yuio.string.AnyString",
            "~yuio.string.NoWrapStart",
            "~yuio.string.NoWrapEnd",
            "~yuio.string.NoWrapMarker",
            "~yuio.json_schema.JsonValue",
            "~yuio.dbg.EnvCollector",
            "~yuio.cli.NArgs",
        ]
        for name in [
            path.rsplit(".", maxsplit=1)[-1].removeprefix("~"),
            path.removeprefix("~"),
        ]
    },
}

vhs_cwd = pathlib.Path(__file__).parent.parent.parent
vhs_min_version = "0.7.2"
vhs_n_jobs_read_the_docs = 4
# vhs_repo = "agentstation/vhs"
# vhs_format = "svg"

toc_object_entries_show_parents = "hide"

# -- Options for HTML output -------------------------------------------------

html_theme = "furo"
html_static_path = ["_static"]
html_css_files = ["extra.css"]
html_js_files = [
    "mermaid-init.js",
    (
        "https://cdn.jsdelivr.net/npm/mermaid@11.4.1/dist/mermaid.min.js",
        {"defer": "defer"},
    ),
]
html_theme_options = {
    "source_repository": "https://github.com/taminomara/yuio",
    "source_branch": "main",
    "source_directory": "docs/source",
}

# fmt: off
from sphinx.util import inspect


# See https://github.com/sphinx-doc/sphinx/issues/14003
class TypeAliasForwardRef(inspect.TypeAliasForwardRef):
    def __repr__(self) -> str:
        return self.name

    def __or__(self, value):
        import typing
        return typing.Union[self, value]

    def __ror__(self, value):
        import typing
        return typing.Union[value, self]

inspect.TypeAliasForwardRef = TypeAliasForwardRef
del TypeAliasForwardRef
del inspect

# See https://github.com/sphinx-doc/sphinx/issues/14005
import sphinx.domains.python._annotations
from sphinx.domains.python._annotations import parse_reftarget as _parse_reftarget


def parse_reftarget(*args, **kwargs):
    _, reftarget, title, refspecific = _parse_reftarget(*args, **kwargs)
    return "obj", reftarget, title, refspecific
sphinx.domains.python._annotations.parse_reftarget = parse_reftarget
# fmt: on



import pathlib
import shutil
import subprocess

import pytest


@pytest.fixture(autouse=True)
def setup_completions_home(tmp_path: pathlib.Path, monkeypatch):
    test_home = pathlib.Path(__file__).parent / "home"

    home = tmp_path / "home"
    shutil.copytree(test_home, home)

    data_home = home / "data"
    data_home.mkdir()

    cache_home = home / "cache"
    cache_home.mkdir()

    config_home = home / "config"
    config_home.mkdir()

    git_template_dir = home / "git_template_dir"
    git_template_dir.mkdir()

    monkeypatch.setenv("HOME", str(home))
    monkeypatch.setenv("XDG_DATA_HOME", str(data_home))
    monkeypatch.setenv("XDG_CACHE_HOME", str(cache_home))
    monkeypatch.setenv("XDG_CONFIG_HOME", str(config_home))
    monkeypatch.setenv("LOCALAPPDATA", str(home))
    monkeypatch.setenv("PATH", str(home), prepend=":")
    monkeypatch.chdir(home)

    monkeypatch.setenv("GIT_CONFIG_GLOBAL", "")
    monkeypatch.setenv("GIT_AUTHOR_NAME", "author")
    monkeypatch.setenv("GIT_AUTHOR_EMAIL", "author@example.com")
    monkeypatch.setenv("GIT_AUTHOR_DATE", "2020-01-01T00:00:00Z")
    monkeypatch.setenv("GIT_COMMITTER_NAME", "committer")
    monkeypatch.setenv("GIT_COMMITTER_EMAIL", "committer@example.com")
    monkeypatch.setenv("GIT_COMMITTER_DATE", "2020-01-02T00:00:00Z")

    subprocess.check_call(
        ["git", "init", ".", "-b", "main", "--template", str(git_template_dir)]
    )
    subprocess.check_call(["git", "commit", "--allow-empty", "-m", "initial commit"])
    subprocess.check_call(["git", "tag", "v1.0.0"])
    subprocess.check_call(["git", "tag", "v1.1.0"])


@pytest.fixture
def test_cases():
    cases = [
        "|",
        "-|",
        "--|",
        "a|",
        "args |",
        "args -|",
        "args --bool|",
        "args --bool=|",
        "args --bool |",
        "args --no-b|",
        "args --int |",
        "args --str |",
        "args --enum |",
        "args --enum op|",
        "args --override |",
        "args --override op|",
        "args --custom |",
        "args --custom c|",
        "args --custom comp_|",
        "args --one-of |",
        "args --git-ref |",
        "args --git-tag |",
        "args --tuple |",
        "args --tuple x |",
        "args --tuple x y |",
        "args --tuple |x y",
        "args --list |",
        "args --list x |",
        "args --list x | y",
        "args --list x --|",
        "args --list x --one-of |",
        "args --list=|",
        "args --list=op|",
        "args --l|ist=op",
        "args --list=x |",
        "args --set |",
        "args --set option_1 |",
        "args --set option_1 o| option_2",
        "args --dict |",
        "args --dict opt|",
        "args --dict oth|",
        "args --dict x:|",
        "args --dict opt|:xxx",
        "args --dict oth|:xxx",
        "args --nested-tuple |",
        "args --nested-tuple opt|",
        "args --nested-tuple oth|",
        "args --nested-tuple x,|",
        "args --nested-tuple x,opt|",
        "args --nested-tuple opt|,x",
        "args --nested-tuple oth|,x",
        "args --nested-list |",
        "args --nested-list opt|",
        "args --nested-list oth|",
        "args --nested-list x,|",
        "args --nested-list x,|,y",
        "args --nested-list x,oth|,y",
        "args --nested-set |",
        "args --nested-set opt|",
        "args --nested-set oth|",
        "args --nested-set x,|",
        "args --nested-set x,|,y",
        "args --nested-set option_1,|,y",
        "args --nested-set x,oth|,y",
        "args --nested-set option_1,y opt|",
        "args --nested-set option_1,y option_1,opt|",
        "args --nested-dict |",
        "args --nested-dict oth|",
        "args --nested-dict x:|",
        "args --nested-dict x:o|",
        "args --nested-dict x:y,|",
        "args --nested-dict x:y,o|ther:zzz",
        "args --union |",
        "args --union oth|",
        "args --enum-by-name |",
        "args --special-chars |",
        "hidden |",
        "hidden --|",
        "hidden --secret |",
        "hidden --secret x |",
    ]

    return list(map(_prepare_test_case, cases))


def _prepare_test_case(args_s: str):
    args = args_s.split()
    cword = 0
    prefix_len = 0
    token_prefix_len = 0
    for i, arg in enumerate(args):
        if i > 0:
            prefix_len += 1
        if "|" in arg:
            cword = i
            l, r = arg.split("|", maxsplit=1)
            prefix_len += len(l)
            token_prefix_len = len(l)
            args[i] = l + r
            break
        else:
            prefix_len += len(arg)
    else:
        assert False, "can't find cursor"
    return cword, prefix_len, token_prefix_len, args, args_s


def extract_results(output: str):
    lines = output.splitlines()
    i = lines.index("--BEGIN RESULTS--")
    j = lines.index("--END RESULTS--")
    res = list(set(filter(lambda x: x, (line.rstrip() for line in lines[i + 1 : j]))))
    res.sort()
    return res



# noqa: INP001
import yuio.app
import yuio.dbg


def collect_env_ok():
    return yuio.dbg.Report(
        title="Collect Env Ok",
        items=[
            "Something something",
            ("foo", "bar"),
        ],
    )


def collect_env_err():
    raise RuntimeError("something went wrong")


@yuio.app.app(version="1.0.0.1")
def main():
    pass



import yuio.app
import yuio.io

@yuio.app.app
def main():
    ALLOWED_VALUES = ["foo", "bar", "baz"]

    yuio.io.info(
        "Allowed values: %s",
        yuio.io.JoinRepr(ALLOWED_VALUES, color="magenta"),
    )

if __name__ == "__main__":
    main.run()



# Yuio project, MIT license.
#
# https://github.com/taminomara/yuio/
#
# You're free to copy this file to your project and edit it for your needs,
# just keep this copyright line please :3

"""
This module provides functionality for collecting data about environment for including
it to bug reports. It's inspired by Scooby__ (check it out if functionality
of this module is not enough).

__ https://github.com/banesullivan/scooby/

Set :attr:`App.bug_report <yuio.app.App.bug_report>` to enable this functionality.

.. autofunction:: print_report

.. autoclass:: ReportSettings
    :members:

.. autoclass:: Report
    :members:

.. type:: EnvCollector
    :canonical: typing.Callable[[], Report]

    Type alias for report collector.

"""

from __future__ import annotations

import contextlib
import dataclasses
import datetime
import os
import re
import sys
import textwrap
import threading
import traceback
import types
from dataclasses import dataclass

import yuio._typing_ext as _tx
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import typing_extensions as _t
else:
    from yuio import _typing as _t

if TYPE_CHECKING:
    import yuio.app

__all__ = [
    "EnvCollector",
    "Report",
    "ReportSettings",
    "print_report",
    "report_exc",
]


@dataclass(slots=True)
class Report:
    """
    An environment report.

    """

    title: str
    """
    Report title.

    """

    items: list[str | tuple[str, str]] = dataclasses.field(default_factory=list)
    """
    Report items. Each item can be a string or a key-value pair of strings.

    """


EnvCollector: _t.TypeAlias = _t.Callable[[], Report]
"""
A callable that collects a report.

"""

_LOCK = threading.Lock()
_ENV_COLLECTORS: list[tuple[str, EnvCollector | Report]] | None = None


@dataclass(slots=True)
class ReportSettings:
    """
    Settings for collecting debug data.

    """

    package: str | types.ModuleType | None = None
    """
    Root package. Used to collect dependencies.

    """

    dependencies: list[str | types.ModuleType] | None = None
    """
    List of additional dependencies to include to version report.

    """

    collectors: list[EnvCollector | Report] | None = None
    """
    List of additional env collectors to run.

    """


def _get_env_collectors():
    global _ENV_COLLECTORS
    if _ENV_COLLECTORS is None:
        with _LOCK:
            if _ENV_COLLECTORS is None:
                _ENV_COLLECTORS = _load_env_collectors()
    return _ENV_COLLECTORS


def _load_env_collectors():
    import importlib.metadata

    collectors: list[tuple[str, EnvCollector | Report]] = []

    for plugin in importlib.metadata.entry_points(group="yuio.dbg.env_collector"):
        try:
            collectors.append((plugin.name, plugin.load()))
        except Exception:
            msg = "Error when loading env collector:\n" + traceback.format_exc()
            collectors.append((plugin.name, Report(plugin.name, [msg])))

    return collectors


@contextlib.contextmanager
def report_exc(report: Report, key: str | None):
    try:
        yield
    except Exception as e:
        if key is not None:
            report.items.append((key, f"can't collect this item: {e}"))
        else:
            report.items.append(f"Can't collect this item: {e}")


def _system() -> Report:
    import platform

    report = Report("System", [])

    now = datetime.datetime.now(datetime.timezone.utc)
    report.items.append(("date", now.strftime("%Y-%m-%d %H:%M:%S")))
    with report_exc(report, "platform"):
        report.items.append(("platform", f"{sys.platform} ({platform.platform()})"))
    with report_exc(report, "os"):
        os = platform.system()
        if os == "Linux":
            data = platform.freedesktop_os_release()
            report.items.append(("os", f"{data['NAME']} {data['VERSION_ID']}"))
        elif os == "Windows":
            report.items.append(("os", " ".join(platform.win32_ver())))
        elif os == "Darwin":
            report.items.append(("os", platform.mac_ver()[0]))
        else:  # pragma: no cover
            report.items.append(("os", os))
    with report_exc(report, "python"):
        report.items.append(
            ("python", f"{platform.python_implementation()} {sys.version}")
        )
    with report_exc(report, "machine"):
        report.items.append(
            ("machine", f"{platform.machine()} {platform.architecture()[0]}")
        )
    return report


def _versions(
    settings: ReportSettings, app: yuio.app.App[_t.Any] | None = None
) -> Report:
    report = Report("Versions")

    package = settings.package
    if app and app.version:
        report.items.append(("__app__", str(app.version)))
    if app and package is None:
        package = app._command.__module__

    dependencies: set[str]
    if package is None:
        dependencies = set()
    elif isinstance(package, (str, types.ModuleType)):
        dependencies = _get_dependencies(package)
    else:
        report.items.append(
            "TypeError: expected str or ModuleType, "
            f"got {_tx.type_repr(type(package))}: {package!r}"
        )
        dependencies = set()

    for dependency in settings.dependencies or []:
        if isinstance(dependency, types.ModuleType):
            dependencies.add(dependency.__name__)
        elif isinstance(dependency, str):
            dependencies.add(dependency)
        else:
            report.items.append(
                "TypeError: expected str or ModuleType, "
                f"got {_tx.type_repr(type(dependency))}: {dependency!r}"
            )
    dependencies.add("yuio")

    for dependency in sorted(dependencies):
        with report_exc(report, dependency):
            report.items.append((dependency, _find_package_version(dependency)))

    return report


def _get_dependencies(package: str | types.ModuleType) -> set[str]:
    import importlib.metadata

    if isinstance(package, types.ModuleType):
        package = package.__name__

    if "[" in package:
        package = package.split("[", maxsplit=1)[0]
    package = package.strip()

    dependencies = {package}
    try:
        distribution = importlib.metadata.distribution(package)
    except importlib.metadata.PackageNotFoundError:
        return dependencies
    requires = distribution.requires
    if not requires:
        return dependencies
    for requirement in requires:
        if match := re.match(r"^\s*([\w-]+)\s*(\[[\w,\s-]*\])?", requirement):
            dependencies.add(match.group(0))
    return dependencies


def _find_package_version(package: str):
    import importlib.metadata

    if "[" in package:
        package = package.split("[", maxsplit=1)[0]
    package = package.strip()

    try:
        return importlib.metadata.version(package)
    except importlib.metadata.PackageNotFoundError:
        pass

    module = importlib.import_module(package)

    for v_string in ("__version__", "version"):
        try:
            version = getattr(module, v_string)
        except AttributeError:
            continue
        if not version:
            continue
        if isinstance(version, str):
            return version
        elif isinstance(version, tuple):
            return ".".join(map(str, version))

    return "unknown"


def _terminal() -> Report:
    import subprocess

    import yuio.io
    import yuio.term

    report = Report("Terminal and CLI")

    report.items.append(("term", os.environ.get("TERM", "")))
    report.items.append(("colorterm", os.environ.get("COLORTERM", "")))
    report.items.append(("shell", os.environ.get("SHELL", "")))
    report.items.append(("ci", repr(yuio.term.detect_ci())))

    if os.name != "nt":  # pragma: no cover
        with report_exc(report, "wsl"):
            try:
                wslinfo = subprocess.getoutput("wslinfo --version")
            except FileNotFoundError:
                wslinfo = "None"
            report.items.append(("wsl", wslinfo))

    report.items.append("Term")

    term = yuio.io.get_term()
    report.items.append(("color support", repr(term.color_support)))
    report.items.append(("ostream", repr(term.ostream)))
    report.items.append(("ostream is tty", repr(term.ostream_is_tty)))
    report.items.append(("istream", repr(term.istream)))
    report.items.append(("istream is tty", repr(term.istream_is_tty)))
    report.items.append(("has terminal theme", repr(term.terminal_theme is not None)))

    report.items.append("TTY")

    yuio.term.get_tty()
    report.items.append(("tty output", repr(getattr(yuio.term, "_TTY_OUTPUT", None))))
    report.items.append(("tty input", repr(getattr(yuio.term, "_TTY_INPUT", None))))
    report.items.append(
        ("explicit colors", repr(getattr(yuio.term, "_EXPLICIT_COLOR_SUPPORT", None)))
    )
    report.items.append(("colors", repr(getattr(yuio.term, "_COLOR_SUPPORT", None))))

    return report


def _collector_name(collector: Report | EnvCollector):
    if isinstance(collector, Report):
        return collector.title
    else:
        qualname = getattr(collector, "__qualname__", "unknown")
        module = getattr(collector, "__module__", "unknown")
        return f"{module}.{qualname}"


def print_report(
    *,
    dest: _t.TextIO | None = None,
    settings: ReportSettings | bool | None = None,
    app: yuio.app.App[_t.Any] | None = None,
):
    """
    Collect and print debug report to the given `dest`.

    :param dest:
        destination stream for printing bug report, default is ``stdout``.
    :param settings:
        settings for bug report generation.
    :param app:
        main app of the project, used to extract project's version and dependencies.

        .. note::

            If your app defined in the ``__main__`` module, Yuio will not be able
            to extract its dependencies and print their versions.

            We recommend defining app in a separate file and importing it to the
            ``__main__`` module.

    """
    if dest is None:
        dest = sys.__stdout__
    if (settings is None or isinstance(settings, bool)) and app is not None:
        settings = app.bug_report
    if settings is None or isinstance(settings, bool):
        settings = ReportSettings()

    all_collectors: list[tuple[str, EnvCollector | Report]] = [
        ("System", _system),
        ("Versions", lambda: _versions(settings, app)),
        ("Terminal and CLI", _terminal),
    ]
    all_collectors.extend(
        (
            (_collector_name(collector), collector)
            for collector in settings.collectors or []
        )
    )
    all_collectors.extend(_get_env_collectors())

    START = 0
    AFTER_TITLE = 1
    AFTER_ITEM = 2
    AFTER_LONG_ITEM = 3

    position = START

    print("```", file=dest)

    col_width = 20
    indent = " " * (col_width + 2)
    for name, collector in all_collectors:
        printed_title = False
        try:
            if isinstance(collector, Report):
                report = collector
            else:
                report = collector()
            title = report.title or name

            if position in [AFTER_ITEM, AFTER_LONG_ITEM]:
                print("\n", file=dest)
            print(indent + title, file=dest)
            print(indent + "~" * len(title) + "\n", file=dest)
            printed_title = True
            position = AFTER_TITLE

            for data in report.items or ["No data"]:
                if isinstance(data, str):
                    key, value = "", data
                else:
                    key, value = data
                if key:
                    if position == AFTER_LONG_ITEM:
                        print("", file=dest)
                    print(f"{key:>{col_width}}: {value}", file=dest)
                    position = AFTER_ITEM
                else:
                    if position in [AFTER_ITEM, AFTER_LONG_ITEM]:
                        print("", file=dest)
                    print(textwrap.indent(value, indent), file=dest)
                    position = AFTER_LONG_ITEM
        except Exception:
            if not printed_title:
                if position in [AFTER_ITEM, AFTER_LONG_ITEM]:
                    print("\n", file=dest)
                print(indent + name, file=dest)
                print(indent + "~" * len(name) + "\n", file=dest)
                position = AFTER_TITLE
            if position in [AFTER_ITEM, AFTER_LONG_ITEM]:
                print("", file=dest)
            print(indent + "Error when collecting information:", file=dest)
            print(
                textwrap.indent(traceback.format_exc(), indent).strip("\n"), file=dest
            )
            position = AFTER_LONG_ITEM

    print("```", file=dest)


if __name__ == "__main__":
    print_report()



from __future__ import annotations

import importlib
import pathlib
import types

import pytest

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import typing_extensions as _t
else:
    from yuio import _typing as _t


root = pathlib.Path(__file__).parent.parent / "yuio"
files = [
    file
    for file in root.glob("*.py")
    if (file.name == "__init__.py" or not file.name.startswith("_"))
]
allowlist = ["commit_id", "version_tuple", "version", "TYPE_CHECKING"]


@pytest.mark.parametrize("file", files, ids=[file.name for file in files])
def test_export_consistency(file: pathlib.Path):
    path = "yuio." + ".".join(file.relative_to(root).parts).removesuffix(".py")
    module = importlib.import_module(path)
    all = set(module.__all__)
    missing = set()
    annotations = _t.get_annotations(module)
    for name, value in module.__dict__.items():
        if name.startswith("_") or name in all or name in allowlist:
            continue
        if isinstance(value, (types.ModuleType, _t.TypeVar, _t.ParamSpec)):
            continue
        if name in annotations:
            missing.add(name)
            continue
        if obj_module := getattr(value, "__module__", ""):
            if not obj_module.startswith("yuio"):
                continue
        missing.add(name)

    if missing:
        raise AssertionError(
            "some items are missing from __all__: " + ", ".join(map(repr, missing))
        )



import time

import yuio.io
import yuio.theme


class BlockProgressTheme(yuio.theme.DefaultTheme):
    # Make spinner faster.
    spinner_update_rate_ms = 100

    msg_decorations = {
        # Custom progress bar symbols.
        "progress_bar/start_symbol": "|",
        "progress_bar/end_symbol": "|",
        "progress_bar/done_symbol": "█",
        "progress_bar/pending_symbol": " ",
        "progress_bar/transition_pattern": "█▉▊▋▌▍▎▏ ",
        # Custom spinner sequence.
        "spinner/pattern": "|||/-\\",
    }

    colors = {
        # Custom accent colors.
        "accent_color": "red",
        "accent_color_2": "magenta",
        # By default, color of the first level heading is `accent_color`;
        # Reset it to `heading_color`.
        "msg/text:heading/1": "heading_color",
        # Decoration and progress bar for tasks.
        "msg/decoration:task": "accent_color",
        "task/progressbar/done/start": "accent_color",
        "task/progressbar/done/end": "accent_color",
    }


if __name__ == "__main__":
    yuio.io.setup(theme=BlockProgressTheme)

    yuio.io.heading("Theme with custom colors and decorations")

    with yuio.io.Task("Doing something") as task:
        time.sleep(4)

        for progress in range(0, 101):
            task.progress(progress / 100)
            time.sleep(0.05)
